#File type: Object: pData
#By Junxiang H., 2023/07/01
#wacmk.com/cn Tech. Supp.

import numpy as np 
class pData: #this is a sample version of Data Object
	def __init__(self, grid, quantities):
		self.grid=grid
		self.quantities=quantities
	def update(self,quantities):
		self.quantities.update(quantities)
grid={ #400 grids
	"x1":np.arange(2,200,(200-2)/512), #20 grids
	"x2":np.arange(np.pi/20,np.pi+np.pi/20,np.pi/280) #20 grids
}

quantities={
	"rho":np.random.random((len(grid["x1"]),len(grid["x2"]))),
	"prs":np.random.random((len(grid["x1"]),len(grid["x2"]))),
	"vx1":np.random.random((len(grid["x1"]),len(grid["x2"]))),
	"vx2":np.random.random((len(grid["x1"]),len(grid["x2"]))),
	"vx3":np.random.random((len(grid["x1"]),len(grid["x2"]))),
	"gamma":4/3,
	"c":3e8,
	"mu":1/2,
	"k":1.38e-23,
	"mp":1.6726231e-27,
	"geometry":"SPHERICAL"
}
try:
	import ShockFinder.Data as Data
	cla=Data.Data
except:
	cla=pData

TestData=cla(grid,quantities)
def get(num):
	return [cla(grid,quantities) for i in range(num)]

if __name__=="__main__":
	print(quantities)


#File type: <Function> set
#By Junxiang H., 2023/07/5
#wacmk.com/cn Tech. Supp.
try:
	import ShockFinder.Addon.Time as Time
except:
	import datetime
	class tt:
		def now():
			return datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")

import h5py,numpy as np,os
Filetype="hdf5"
def PreSave(filename,*quantities_name,**databaseinfo):
	if "pnum" in databaseinfo.keys():
		pnum=databaseinfo["pnum"]
	else:
		pnum=1
	tfilename=databaseinfo["projectname"]
	ssfname=tfilename
	if "/" in tfilename:
		ssfname=tfilename.split("/")[-1]
	if "\\" in tfilename:
		ssfname=tfilename.split("\\")[-1]
	if not os.path.exists(tfilename):
		os.makedirs(tfilename)
	for i in range(0,pnum):
		subfilename=os.path.join(tfilename,ssfname+str(i)+"."+Filetype)
		subfile=h5py.File(subfilename,"w")
		subfile.create_group("Data")
		subfile.close()
	hdffile=h5py.File(filename,"w")
	for i in databaseinfo.keys():
		try:
			hdffile.attrs[i]=databaseinfo[i]
		except:
			hdffile.attrs[i]=str(databaseinfo[i])
	hdffile.attrs["Time"]=Time.now()
	hdffile.attrs["HDF5LABEL"]="ShockFinderDataFile"
	hdffile.close()
def Save(filename,data,index,*quantities_name):
	hdffile=h5py.File(filename,"a")
	g_dt=hdffile["Data"]
	try:
		g_dti=g_dt.create_group(str(index))
		g_dtig=g_dti.create_group("Grid")
		g_dtiq=g_dti.create_group("Quantities")
	except Exception as err:
		g_dti=g_dt[str(index)]
		g_dtig=g_dti["Grid"]
		g_dtiq=g_dti["Quantities"]
	for i in quantities_name:
		try:
			g_dtiq.create_dataset(i,data=data.quantities[i])
		except Exception as err1:
			try:
				g_dtig.create_dataset(i,data=data.grid[i])
			except Exception as err2:
				pass
				#print("Error:",err1,err2)
	hdffile.close()
try:
	import ShockFinder.DataBase as DataBase
	import ShockFinder.Data as Data
	def Load(filename,MPE=None,print=print,**vargs):
		filename=filename.rsplit("."+Filetype,1)[0]
		hdffile=h5py.File(filename+"."+Filetype,"r")
		dbinfo={}
		for i in hdffile.attrs.keys():
			dbinfo[i]=hdffile.attrs[i]
			if dbinfo[i]=="None":
				dbinfo[i]=None
		if "HDF5LABEL" not in hdffile.attrs or hdffile.attrs["HDF5LABEL"] != "ShockFinderDataFile":
			#print("Error: \"",filename,"\" is not a ShockFinder data file")
			return None
		hdffile.close()
		if MPE==None:
			return DataBase.DataBaseConnect(*[load_fun(dbinfo["projectname"],i,dbinfo,print) for i in range(dbinfo["pnum"])])
		#MPE mode
		mpe=MPE.build(load_fun,**vargs)
		return DataBase.DataBaseConnect(*mpe.fun(dbinfo["projectname"],mpe.Array(range(dbinfo["pnum"])),dbinfo))
except:
	def Load(filename,MPE=None,**vargs):
		print("Load module build failure, please use orther IO modules")
def load_fun(projectname,projectindex,dbinfo,print=print):
	ndb=DataBase.DataBase(infomation={},printfun=print)
	ssfname=projectname
	if "/" in projectname:
		ssfname=projectname.split("/")[-1]
	if "\\" in projectname:
		ssfname=projectname.split("\\")[-1]
	subfile=os.path.join(projectname,ssfname+str(projectindex)+"."+Filetype)
	hdffile=h5py.File(subfile,"r")
	g_dt=hdffile["Data"]
	for i in g_dt.keys():
		print("Loading \"",i,"\" in \"",subfile,"\"")
		grid={}
		g_dtg=g_dt[i]["Grid"]
		for j in g_dtg.keys():
			grid[j]=np.array(g_dtg[j])
		quantities={}
		g_dtq=g_dt[i]["Quantities"]
		for j in g_dtq.keys():
			quantities[j]=np.array(g_dtq[j])
			if quantities[j].ndim==0:
				try:
					quantities[j]=float(quantities[j])
				except:
					quantities[j]=quantities[j].item().decode("utf-8")
					if quantities[j]=="True":
						quantities[j]=True
					elif quantities[j]=="False":
						quantities[j]=False
					elif "," in quantities[j]:
						quantities[j]=quantities[j].split(",")
		ndb.data.append(Data.Data(grid,quantities))
		ndb.tindex.append(int(i))
	ndb.update_infomation(**dbinfo)
	hdffile.close()
	ndb.data=tuple(ndb.data)
	ndb.tindex=tuple(ndb.tindex)
	print("Loading finished, waiting for database connection")
	return ndb


#File type: extension <Function> set
#By Junxiang H., 2023/06/30
#wacmk.com/cn Tech. Supp.

import os,re
#return Filesdir/Filename.type
def GetFiles(filedir=os.getcwd(),filetype=""):
	result=()
	'''
	for root, dirs,files in os.walk(filedir,topdown=False):
		for file in files:
			if file.endswith(filetype):
				result+=(os.path.join(root,file),)
	'''
	for file in os.listdir(filedir):
		if file.endswith(filetype):
			result+=(os.path.join(filedir,file),)
	#result = tuple(sorted(result, key=lambda x: (int(re.sub('\D', '', x)), x)))
	return result

if __name__=="__main__":
	print("Testing Model:",__file__)
	print("Testing Function:", GetFiles)
	print("Testing Result:", GetFiles(filetype="py"))


#File type: extension <class page> set
#By Junxiang H., 2023/07/9
#wacmk.com/cn Tech. Supp.

#This script updates automaticly! Do not Modify!
#Update time:2023-10-22 14:55:09

Pages={}
import ShockFinder.Addon.GUI.XUIlib.pages.Help as Help
Pages["Help"]=Help
import ShockFinder.Addon.GUI.XUIlib.pages.Figure as Figure
Pages["Figure"]=Figure
import ShockFinder.Addon.GUI.XUIlib.pages.Analyze as Analyze
Pages["Analyze"]=Analyze
import XenonUI.XUIlib.pages.Exit as Exit
Pages["Exit"]=Exit



#File type: <Function> return a new <Object: Data>
#By Junxiang H., 2023/07/01
#wacmk.com/cn Tech. Supp.

#if you would like to import some packages,
#during the data loading.
#Please put that packages into this folder and using:

try:
	import ShockFinder.Addon.AnalysisTool.Mean as Mean
	from ShockFinder.Addon.AnalysisTool.Basic import *
except:
	import Mean #debug
	from Basic import *
import copy
import numpy as np
need=["rho", "prs","c","mu","mp","k"]
def get(Dataobj,args={},vargs={}):
	Dataobj.quantities.update(args)
	for i in need:
		if i not in Dataobj.quantities.keys() and i not in vargs.keys():
			print("Warning: args:",i,"is needed without definding")
			return Dataobj
	quantities={"Temperature":Dataobj.quantities["prs"]/Dataobj.quantities["rho"]*Dataobj.quantities["c"]**2*Dataobj.quantities["mu"]*Dataobj.quantities["mp"]/Dataobj.quantities["k"]}
	Dataobj.quantities.update(quantities)
	if "AvgTh" in vargs.keys() and vargs["AvgTh"]==True:
		Dataobj=get_AvgTh(Dataobj,args,vargs)
	return Dataobj
def result(quantity_name=None,anafname=None):
	return ("Temperature","AvgTh_Temperature")

AvgTh_cal=True
def get_AvgTh(Dataobj,args={},vargs={"Mean_axis":(1,)}):
	try:
		if AvgTh_cal:
			newneed=copy.deepcopy(need)
			Dataobj.quantities.update(args)
			for i in newneed:
				if i not in Dataobj.quantities.keys() and i not in vargs.keys():
					print("Warning: args:",i,"is needed without definding")
					return Dataobj
			meanstr=""
			meanaxis=get_par(Dataobj,vargs,"Mean_axis",Dataobj.quantities["rho"].ndim-1)
			if type(meanaxis) not in (list,tuple,np.ndarray):
				meanaxis=(meanaxis,)
			for i in meanaxis:
				meanstr+=str(i)+"@"+str((round(vargs["Mean_axis"+str(i)][0],2),round(vargs["Mean_axis"+str(i)][1],2)))+"_" if "Mean_axis"+str(i) in vargs.keys() else str(i)+"@"+str((round(Dataobj.quantities["Mean_axis"+str(i)][0],2),round(Dataobj.quantities["Mean_axis"+str(i)][1],2)))+"_" if "Mean_axis"+str(i) in Dataobj.quantities.keys() else ""
			quantities={"AvgTh_"+meanstr+"Temperature":Mean.Mean(Dataobj,"prs",vargs)/Mean.Mean(Dataobj,"rho",vargs)*Dataobj.quantities["c"]**2*Dataobj.quantities["mu"]*Dataobj.quantities["mp"]/Dataobj.quantities["k"]} #update here
			Dataobj.quantities.update(quantities)
		else:
			print("Warning: AvgTh mode is not opened:", "Temperature") #update here
	except:
		print("Warning: AvgTh mode is not definded or error:", __file__) #update here
	return Dataobj

if __name__=="__main__":
	print("Testing Model:",__file__)
	from TestData import TestData
	TestData=get(TestData)
	print("Testing Result:", TestData.quantities["Temperature"])
	if AvgTh_cal:
		TestData=get_AvgTh(TestData)
		print("Testing Result:", TestData.quantities["AvgTh_Temperature"])


#File type: <Function> set
#By Junxiang H., 2023/07/03
#wacmk.com/cn Tech. Supp.

try:
	import ShockFinder.Addon.Painter.Basic as Basic
except:
	import Basic

from matplotlib import pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
import pandas as pd,numpy as np
def set_figure_info(ax,**Figureinfo):
	title=Basic.CharUndecode(Basic.get_par(Figureinfo,"title",""))
	x_axis=Basic.CharUndecode(Basic.get_par(Figureinfo,"x_axis",""))
	y_axis=Basic.CharUndecode(Basic.get_par(Figureinfo,"y_axis",""))
	z_axis=Basic.CharUndecode(Basic.get_par(Figureinfo,"z_axis",""))
	x_lim=Basic.get_par(Figureinfo,"x_lim")
	y_lim=Basic.get_par(Figureinfo,"y_lim")
	z_lim=Basic.get_par(Figureinfo,"z_lim")
	if x_lim!=None:
		ax.set_xlim(*x_lim)
	if y_lim!=None:
		ax.set_ylim(*y_lim)
	if z_lim!=None:
		ax.set_zlim(*z_lim)
	ax.set_title(title)
	ax.set_xlabel(x_axis)
	ax.set_ylabel(y_axis)
	ax.set_zlabel(z_axis)
	
def line(*lines,**Figureinfo):
	showlabel=False
	fig=plt.figure()
	ax=Axes3D(fig,auto_add_to_figure=False)
	fig.add_axes(ax)
	set_figure_info(ax,**Figureinfo)
	for line in lines:
		le,li=line
		ax.plot(le["x"],le["y"],le["z"],**li)
		if Basic.get_par(li,"label")!=None:
			showlabel=True
	if showlabel:
		ax.legend()
	plt.show()


def surface(surface,**Figureinfo):#
	fig=plt.figure()
	ax=Axes3D(fig,auto_add_to_figure=False)
	fig.add_axes(ax)
	set_figure_info(ax,**Figureinfo)
	sf,sfi=surface
	x,y=np.meshgrid(sf["x"],sf["y"])
	suf=ax.plot_surface(x.T,y.T,sf["v"],**sfi)
	if str(Basic.get_par(Figureinfo,"contour",False))=="True":
		ax.contour(x.T,y.T,sf["v"],zdir='z',offset=-1,cmap=plt.get_cmap("rainbow"))
	plt.colorbar(suf)
	plt.show()

def scatter(scatter,**Figureinfo):
	fig = plt.figure()
	ax = fig.add_subplot(111, projection='3d')
	set_figure_info(ax,**Figureinfo)
	sc,sci=scatter
	sci["facecolors"]=plt.cm.jet(sc["v"])
	surf = ax.plot_surface(sc["x"], sc["y"], sc["z"], **sci)
	plt.colorbar(surf, orientation='vertical')
	plt.show()


def info():
	print("Module:",__file__)

if __name__=="__main__":
	import Line,Surface,math
	'''
	x=np.arange(100)
	y=np.arange(0,2*np.pi,2*0.01*np.pi)
	z=np.sin(y)/x
	l=Line.CreateLine(x=x,y=y,z=z,label="$\\frac{sin(y)}{x}$")
	line(l,label=True,title="Test")
	
	'''
	r=np.arange(50)
	t=np.arange(0,(1+0.02)*np.pi,1*0.02*np.pi)
	p=np.arange(0,(2+0.02)*np.pi,2*0.02*np.pi)
	v=r.reshape(50,1,1)*np.sin(t).reshape(1,51,1)*np.cos(p).reshape(1,1,51)
	x,y,z,v=Basic.rtp_to_xyz(r,t,p,v)
	cx=np.full((len(r),),10)
	cy=r
	cz=r
	x,y,v=Basic._3D_to_2D(x,y,z,v,cx,cy,cz)
	#x,y,v=Basic.rop_to_xoy(r,p,v)
	#print(v)
	sf=Surface.CreateSurface(x=x,y=y,v=v,cmap="hot")
	surface(sf,x_axis="x",y_axis="y",z_axis="z")
	


#File type: <Function> return <Object: Data>
#By Junxiang H., 2023/07/1
#wacmk.com/cn Tech. Supp.

try:
	import ShockFinder.Addon.Loader.Pluto_lib.pload as pp
except:
	import Pluto_lib.pload as pp

try:
	import ShockFinder.Addon.Loader.FileNamePreProcess as FNPP
except:
	import FileNamePreProcess as FNPP

try:
	import ShockFinder.Data as Data
except:
	pass

def load(filename): 
	#Loading Process
	fileindex,filedir,filetype=FNPP.FileNamePreProcess(filename)
	import os
	filedir=os.path.join(filedir,"dbl.out").rsplit("dbl.out")[0]
	read=pp.pload(int(fileindex),w_dir=filedir,datatype=filetype) 
	grid={}
	#GEOMETRY:	SPHERICAL	CYLINDRICAL		POLAR					CARTESIAN
	#			x1-x2-x3	x1-x2			x1-x2-x3				x1-x2-x3
	#			r-theta-phi r-z				r-phi(or theta)-z		x-y-z
	grid["x1"]=read.x1 #updated here
	grid["x2"]=read.x2 #updated here
	grid["x3"]=read.x3 #updated here
	#basic quantities
	quantities={}
	#user definded...
	try: #vector model
		quantities["vx1"]=read.vx1
	except:
		pass
	try: #vector model
		quantities["vx2"]=read.vx2
	except:
		pass
	try: #vector model
		quantities["vx3"]=read.vx3
	except:
		pass
	quantities["SimTime"]=read.SimTime
	quantities["geometry"]=read.geometry
	quantities["rho"]=read.rho
	quantities["prs"]=read.prs
	try: #vector model
		quantities["Bx1"]=read.Bx1
	except:
		pass
	try: #vector model
		quantities["Bx2"]=read.Bx2
	except:
		pass
	try: #vector model
		quantities["Bx3"]=read.Bx3
	except:
		pass
	try:
		dt=Data.Data(grid,quantities)
		return dt
	except:
		return (grid,quantities)


from msilib.schema import File
try:
	from ShockFinder.Config import * 
except:
	Config={}
	Default={}
from ShockFinder.DataBase import DataBase
from ShockFinder.Update import Update
import numpy as np,copy,os
from multiprocessing import cpu_count
def args_cut(args,pnum):
	result=[]
	for i in range(pnum):
		result.append([])
	cal_num=len(args)
	pp=0
	for i in args:
		result[pp].append(i)
		pp+=1
		if pp==pnum:
			pp=0
	return result
def test_database(num,tindex=None,printfun=print):
	db=DataBase(Default,printfun)
	db.build_test_data(AnalysisLib["TestData"].get,num)
	db.tindex=tindex if tindex!=None else range(num)
	return db
def analyze(analyze_commands,filesname={},testmode=False,print=print):
	if testmode:
		num=len(filesname)
		database=test_database(num,list(filesname.keys()),print)
	else:
		database=DataBase(infomation=Default,printfun=print)
		database.load_data(analyze_commands["Loader"],**filesname)
	quantities_name=[]
	for i in analyze_commands.keys():
		if i=="Update":
			database.update(**analyze_commands[i])
		elif i=="Test" and analyze_commands[i]:
			#all tools
			vargs={"Mean_axis":(1,),"Radial_axis":(1,),"Radial_value":(None,),"ShockDeduct":0.1,"Harmonic_Max_l":3,"Harmonic_Max_m":3}
			#if "IO" in analyze_commands.keys():
			#	vargs.update(analyze_commands["IO"])
			#standard
			standard=("SoundSpeed","Temperature","Entropy","MachNumber","Bremsstrahlung","MassFlux")
			quantities_name=("rho","prs","vx1","vx2","vx3")+standard
			ext_tools=("Harmonic","Mean","Gradient","Divergence","Radial")
			for j in standard:
				nvgs=copy.deepcopy(vargs)
				nvgs.update({"info":j,"result":j})
				anaf=AnalysisTool[j].get
				database.analysis_data(anaf,**nvgs)
				if not database.check_quantities(j):
					exit()
			for j in quantities_name:
				for k in ext_tools:
					nvgs=copy.deepcopy(vargs)
					nvgs.update({"info":k,"quantity_name":j})
					if k not in ("Harmonic","Gradient"):
						nvgs.update({"result":k+"_"+j})
					elif k=="Harmonic" and database.get_geometry()=="SPHERICAL":
						rrs=[]
						for l in range(vargs["Harmonic_Max_l"]+1):
							if database.get_dimension()==2:
								rrs.append(k+"_l@"+str(l)+"_"+j)
							if database.get_dimension()==3:
								for m in range(max(-vargs["Harmonic_Max_l"],-l),min(vargs["Harmonic_Max_l"]+1,l+1)):
									rrs.append(k+"_l@"+str(l)+"_m@"+str(m)+"_"+j)
						nvgs.update({"result":rrs})
					elif k=="Gradient":
						if database.get_dimension()==2:
							nvgs.update({"result":[k+"_"+j+"_x1",k+"_"+j+"_x2"]})
						elif database.get_dimension()==3:
							nvgs.update({"result":[k+"_"+j+"_x1",k+"_"+j+"_x2", k+"_"+j+"_x3"]})
					anaf=AnalysisLib[k]
					database.analysis_data(anaf,**nvgs)
					if k not in ("Harmonic","Gradient"):
						if not database.check_quantities(k+"_"+j):
							exit()
					elif k=="Harmonic" and database.get_geometry()=="SPHERICAL":
						for l in range(vargs["Harmonic_Max_l"]+1):
							if database.get_dimension()==2 and not database.check_quantities(k+"_l@"+str(l)+"_"+j):
								exit()
							if database.get_dimension()==3:
								for m in range(max(-vargs["Harmonic_Max_l"],-l),min(vargs["Harmonic_Max_l"]+1,l+1)):
									if not database.check_quantities(k+"_l@"+str(l)+"_m@"+str(m)+"_"+j):
										exit()
					elif k=="Gradient":
						if database.get_dimension()==2 and not database.check_quantities([k+"_"+j+"_x1",k+"_"+j+"_x2"]) or database.get_dimension()==3 and not database.check_quantities([k+"_"+j+"_x1",k+"_"+j+"_x2", k+"_"+j+"_x3"]):
							exit()
				nvgs=copy.deepcopy(vargs)
				nvgs.update({"info":"Shocks","quantity_name":j})
				anaf=AnalysisTool["Shocks"].get
				database.analysis_data(anaf,**nvgs)
				if not database.check_quantities("Shock_"+j):
					exit()
		elif i not in ("IO","Loader"):
			quantities_name.append(i)
			analyze_commands[i].update({"info":i})
			if "IO" in analyze_commands.keys():
				analyze_commands[i].update(analyze_commands["IO"])
			if "result" not in analyze_commands[i].keys():
				pass
				#analyze_commands[i].update({"result":i})
			anaf=None
			try:
				anaf=AnalysisLib[i]
			except:
				try:
					anaf=AnalysisTool[i].get
				except:
					continue
			database.save_analyze_pars(anaf,**analyze_commands[i])
	database.analyze()
	print("Finished analyzing!")
	database.data=tuple(database.data)
	database.tindex=tuple(database.tindex)
	return database
def get_par(args,name,default=None):
	try:
		return args[name]
	except:
		return default
class ShockFinder:
	Default_MPE="XME"
	Default_IO="HDF5"
	Default_GUI="XUI"
	try:
		MultiprocessEngine=MultiprocessEngine[Default_MPE]
	except:
		MultiprocessEngine=None
	try:
		IO=IO[Default_IO]
	except:
		IO=None
	try:
		GUI=GUI[Default_GUI]
	except:
		GUI=None
	Config=Config
	MPE=None
	database=DataBase({})
	testdb=test_database(1,[0])
	testmode=False
	def __init__(self,config_file=None):
		if config_file!=None: #MPA MODE
			self.command(GeneralLib["ConfigReader"].get_config(config_file))
		else:#GUI MODE
			#set default load info
			self.LMPEINFO={ }
			self.LMPEINFO["cpu_leave"]=1 #left 1 maintains system stable
			self.LMPEINFO["cpu_logicmax"]=4 #set a maximul
			self.LMPEINFO["pnum"]=cpu_count()-self.LMPEINFO["cpu_leave"] if cpu_count()>self.LMPEINFO["cpu_leave"] and cpu_count()<=self.LMPEINFO["cpu_logicmax"] else self.LMPEINFO["cpu_logicmax"] if cpu_count()>self.LMPEINFO["cpu_logicmax"] else self.LMPEINFO["cpu_leave"] 
			if self.GUI!=None:
				self.GUIobj=self.GUI.show(**{"Infobj":self,"windows_title":"ShockFinder "+str(ShockFinderVersion)})
				self.testdb.printfun=self.GUIobj.iofun
				self.database.printfun=self.GUIobj.iofun # is not needed in logical
				self.GUIobj.iofun("Wellcome to use ShockFinder ^_^",color="green")
				self.GUIobj.guiobj.show()
				#self.GUI.CollectPages()
	def set_MPE(self,mpe):
		try:
			self.MultiprocessEngine=MultiprocessEngine[mpe]
		except Exception as err:
			try:
				self.GUIobj.iofun(err,color="red")
			except:
				print(err)
	def set_IO(self,io):
		try:
			self.IO=IO[io]
		except Exception as err:
			try:
				self.GUIobj.iofun(err,color="red")
			except:
				print(err)
	def load(self,filename):
		try:
			if self.IO!=None:
				tmpdb=self.IO.Load(filename,self.MultiprocessEngine,**self.LMPEINFO)
				if type(tmpdb)==type(None):
					self.GUIobj.iofun("Loading error:",filename,"is not a ShockFinder data file",color="red")
					return False
				self.database=tmpdb
				return True
			self.GUIobj.iofun("No IO module selected",color="blue")
			return False
		except Exception as err:
			try:
				self.GUIobj.iofun("Load database",filename," failure! (Error:",err,")",color="red")
			except:
				print(err)
			return False

	def command(self,command):
		analyze_commands={}
		filesname={}
		for i in command.keys():
			values=command[i]
			if i=="MultiprocessEngine":
				self.MultiprocessEngine=MultiprocessEngine[values["Engine"]]
				self.MPE=self.MultiprocessEngine.build(analyze,**values)
			elif i=="IO":
				self.IO=IO[values["Engine"]]
				qtname=[]
				for ii in command.keys():
					qtname+=ii if ii not in ("MultiprocessEngine","IO","Testdata","Update","Test") else ""
				databaseinfo=self.testdb.infomation
				if "Loader" in command.keys():
					databaseinfo.update({
							"FileDir":get_par(command["Loader"],"FileDir",os.getcwd()),
							"FileType":get_par(command["Loader"],"FileType","dbl"),
							"Interval":get_par(command["Loader"],"Interval",1)
						})
				else:
					databaseinfo.update({"Interval":1})
				tfname=get_par(values,"filename",str(int(GeneralLib["Time"].get_time())))
				filename=tfname+"."+self.IO.Filetype

				databaseinfo.update({"pnum":1 if self.MPE==None else self.MPE.pnum,"projectname":tfname})
				self.IO.PreSave(filename,*qtname,**databaseinfo)
				analyze_commands["IO"]=values
				analyze_commands["IO"].update({"IOfun":self.IO.Save,"filename":filename,"projectname":tfname})
			elif i=="Loader":
				analyze_commands[i]=Loader[values["Engine"]].load
				allfiles=LoaderLib["FPP"].GetFiles(get_par(values,"FileDir",os.getcwd()),get_par(values,"FileType","dbl"))
				if self.MPE!=None:
					TrueIndex=args_cut(range(0,len(allfiles),get_par(values,"Interval",1)),self.MPE.pnum)
					filesname=[{str(j):allfiles[j] for j in TrueIndex[i]} for i in range(self.MPE.pnum)]
				else:
					TrueIndex=range(0,len(allfiles),get_par(values,"Interval",1))
					for j in TrueIndex:
						filesname[str(j)]=allfiles[j] 
			elif i=="Testdata":
				self.testmode=True
				if self.MPE!=None:
					TrueIndex=args_cut(range(values["num"]),self.MPE.pnum)
					filesname=[{j:j for j in TrueIndex[i]} for i in range(self.MPE.pnum)]
				else:
					for i in range(values["num"]):
						filesname[i]=i 
				#self.use_test_data(*num)
			else:
				analyze_commands[i]=values
		self.analyze(analyze_commands,filesname)
	def analyze(self,analyze_commands,filesname): #
		if self.MPE!=None:
			anacmd=[]
			for i in range(self.MPE.pnum):
				anacmd.append(copy.deepcopy(analyze_commands))
				tfilename=anacmd[i]["IO"]["projectname"]
				ssfname=tfilename
				if "/" in tfilename:
					ssfname=tfilename.split("/")[-1]
				if "\\" in tfilename:
					ssfname=tfilename.split("\\")[-1]
				if anacmd[i]["IO"]["Engine"]=="HDF5":
					anacmd[i]["IO"]["filename"]=os.path.join(tfilename,ssfname+str(i)+"."+self.IO.Filetype)
			self.MPE.fun(self.MPE.Array(anacmd),self.MPE.Array(filesname),self.testmode)
		else:
			analyze(analyze_commands,filesname)



#File type: <Function> return a new <Object: Data>
#By Junxiang H., 2023/07/01
#wacmk.com/cn Tech. Supp.

#if you would like to import some packages,
#during the data loading.
#Please put that packages into this folder and using:

need=["vx1","vx2","vx3","rho"]
import numpy as np
#args will be inserted into Data Object
#vargs will not be inserted into Data Object
def get(Dataobj,args={},vargs={}):
	Dataobj.quantities.update(args)
	for i in need:
		if i not in Dataobj.quantities.keys() and i not in vargs.keys():
			print("Warning: args:",i,"is needed without definding")
			return Dataobj
	quantities={
		#operation with dict args
		#...
	}
	for i in range(Dataobj.quantities["rho"].ndim):
		quantities["MassFlux_x"+str(i+1)]=Dataobj.quantities["rho"]*Dataobj.quantities["vx"+str(i+1)]
	Dataobj.quantities.update(quantities)
	return Dataobj
def result(quantity_name=None,anafname=None):
	return ("MassFlux_x1","MassFlux_x2","MassFlux_x3")

if __name__=="__main__":
	print("Testing Model:",__file__)
	from TestData import TestData
	TestData=get(TestData)
	print("Testing Result:MassFlux_x1", TestData.quantities["MassFlux_x1"])


#This is a model file for the Loader Addon
#It will be created when creating a new Loader
#Note!!!!: places which marke "<>" have to be updated, and delete "<>".

#File type: <Function> return <Object: Data>
#By Junxiang H., 2023/06/30
#wacmk.com/cn Tech. Supp.

#var filename includes the file_dir

#if you would like to import some packages,
#during the data loading.
#Please put that packages into this folder and using:

'''
try:
	import ShockFinder.Addon.Loader.<package1name> as <package1name>
	import ShockFinder.Addon.Loader.<package2name> as <package2name>
	#...
except:
	import <package1name> #debug
	import <package2name> #debug
	#...
'''


#A default Loader Addon can preprocess the filename,
#into to a formation with (time_index, file_dir, file_type)
#You can denote the below sentence to use it.

'''
try:
	import ShockFinder.Addon.Loader.FileNamePreProcess.FileNamePreProcess as FNPP
except:
	import FileNamePreProcess.FileNamePreProcess #debug
'''

try:
	import ShockFinder.Data
except:
	pass
def load(filename): #updated here
	#Loading Process
	read=<reader Process>(filename) #updated here
	#grid definded
	grid={}
	#GEOMETRY:	SPHERICAL	CYLINDRICAL		POLAR		CARTESIAN
	#			x1-x2-x3	x1-x2			x1-x2-x3	x1-x2-x3
	#			r-theta-phi r-z				r-phi-z		x-y-z
	grid["x1"]=read.<x1> #updated here
	grid["x2"]=read.<x2> #updated here
	grid["x3"]=read.<x3> #updated here 
	#basic quantities
	quantities={}
	quantities["vx1"]=read.<vx1> #updated here
	quantities["vx2"]=read.<vx2> #updated here
	quantities["vx3"]=read.<vx3> #updated here
	quantities["rho"]=read.<rho> #updated here
	quantities["prs"]=read.<prs> #updated here
	quantities["geometry"]=read.<geometry>
	#user definded...
	try:
		return Data.Data(grid,quantities)
	except:
		return (grid,quantities)


#File type: <Function> set
#By Junxiang H., 2023/07/4
#wacmk.com/cn Tech. Supp.

def str_clean(strc):
	if strc[-1]=="\n":
		strc=strc[:-1]
	result=""
	while len(strc)>1 and strc[0]==" ":
		strc=strc[1:]
	for i in ("#","//"):
		strc=strc.split(i)[0] 
	stat=False
	for i in strc:
		if i==" ":
			if not stat:
				stat=True
				result+=i
		else:
			stat=False
			result+=i
	return result if result!="" else None
def retype_string(string):
	if "." in string or "e" in string:
		try:
			string=float(string)
		except:
			if string=="None":
				string=None
			elif string=="True":
				string=True
			elif string=="False":
				string=False
	else:
		try:
			string=int(string)
		except:
			pass
	return string if type(string) !=str or len(string.split(","))==1 else [retype_string(i) for i in string.split(",")]

def get_config(config_file):
	file=open(config_file,"r")
	commands={}
	for line in file.readlines():
		result=str_clean(line)
		if result!=None:
			if len(result.split("@"))>=2:#type2
				#type 1:
				#key=values
				#type 2: key as command
				#command@key1=value1@key2=value2@...
				tt=result.split("@")
				key=tt[0]
				subcommands={}
				for subcmd in tt[1:]:
					if subcmd!="":
						subkey,value=subcmd.split("=")
						subcommands[subkey]=retype_string(value)
				commands[key]=subcommands
			else:#type1
				if len(result.split("="))==2:
					key,value=result.split("=")
					commands[key]=retype_string(value)
				else:
					print("Unknow configuration:",result)
	file.close()
	return commands


from __future__ import division
import numpy as np
import scipy.ndimage
from scipy.interpolate import interp1d, UnivariateSpline

class Tools(object):
	"""
		This Class has all the functions doing basic mathematical
		operations to the vector or scalar fields.
		It is called after pyPLUTO.pload object is defined.
	"""
	def find(self,condition):
		res, = np.nonzero(np.ravel(condition))
		return res
		
	def deriv(self,Y,X=None):
		"""
			Calculates the derivative of Y with respect to X.
				
			**Inputs:**
			Y : 1-D array to be differentiated.\n
			X : 1-D array with len(X) = len(Y).\n
				
			If X is not specified then by default X is chosen to be an equally spaced array
			having same number of elements as Y.
			**Outputs:**
			This returns an 1-D array having the same no. of elements as Y (or X) and
			contains the values of dY/dX.
		"""
	
		n = len(Y)
		n2 = n-2
		if X==None : X = np.arange(n)
		Xarr = np.asarray(X,dtype='float')
		Yarr = np.asarray(Y,dtype='float')
		x12 = Xarr - np.roll(Xarr,-1)   #x1 - x2
		x01 = np.roll(Xarr,1) - Xarr    #x0 - x1
		x02 = np.roll(Xarr,1) - np.roll(Xarr,-1) #x0 - x2
		DfDx = np.roll(Yarr,1) * (x12 / (x01*x02)) + Yarr * (1./x12 - 1./x01) - np.roll(Yarr,-1) * (x01 / (x02 * x12))
		# Formulae for the first and last points:
		
		DfDx[0] = Yarr[0] * (x01[1]+x02[1])/(x01[1]*x02[1]) - Yarr[1] * x02[1]/(x01[1]*x12[1]) + Yarr[2] * x01[1]/(x02[1]*x12[1])
		DfDx[n-1] = -Yarr[n-3] * x12[n2]/(x01[n2]*x02[n2]) + Yarr[n-2]*x02[n2]/(x01[n2]*x12[n2]) - Yarr[n-1]*(x02[n2]+x12[n2])/(x02[n2]*x12[n2])
		
		return DfDx


	def Grad(self,phi,x1,x2,dx1,dx2,polar=False):
		""" This method calculates the gradient of the 2D scalar phi.
			
			**Inputs:**
			
			phi -- 2D scalar whose gradient is to be determined.\n
			x1 -- The 'x' array\n
			x2 -- The 'y' array\n
			dx1 -- The grid spacing in 'x' direction.\n
			dx2 -- The grid spacing in 'y' direction.\n
			polar -- The keyword should be set to True inorder to estimate the Gradient in polar co-ordinates. By default it is set to False.
			
			**Outputs:**
			
			This routine outputs a 3D array with shape = (len(x1),len(x2),2), such that [:,:,0] element corresponds to the gradient values of phi wrt to x1 and [:,:,1] are the gradient values of phi wrt to x2.
			
		"""
		(n1, n2) = phi.shape
		grad_phi = np.zeros(shape=(n1,n2,2))
		h2 = np.ones(shape=(n1,n2))
		if polar == True:
			for j in range(n2):
					h2[:,j] = x1
							
		for i in range(n1):
				scrh1 = phi[i,:]
				grad_phi[i,:,1] = self.deriv(scrh1,x2)/h2[i,:]
		for j in range(n2):
				scrh2 = phi[:,j]
				grad_phi[:,j,0] = self.deriv(scrh2,x1)
		return grad_phi
			
	def Div(self,u1,u2,x1,x2,dx1,dx2,geometry=None):
		"""
			This method calculates the divergence of the 2D vector fields u1 and u2.
		
		**Inputs:**
		
		u1 -- 2D vector along x1 whose divergence is to be determined.\n
		u2 -- 2D vector along x2 whose divergence is to be determined.\n
		x1 -- The 'x' array\n
		x2 -- The 'y' array\n
		dx1 -- The grid spacing in 'x' direction.\n
		dx2 -- The grid spacing in 'y' direction.\n
		geometry -- The keyword *geometry* is by default set to 'cartesian'. It can be set to either one of the following : *cartesian*, *cylindrical*, *spherical* or *polar*. To calculate the divergence of the vector fields, respective geometric corrections are taken into account based on the value of this keyword.
		
		**Outputs:**
		
		A 2D array with same shape as u1(or u2) having the values of divergence.
		"""
		(n1, n2) = u1.shape
		Divergence = np.zeros(shape=(n1,n2))
		du1 = np.zeros(shape=(n1,n2))
		du2 = np.zeros(shape=(n1,n2))
		
		A1 = np.zeros(shape=n1)
		A2 = np.zeros(shape=n2)
		
		dV1 = np.zeros(shape=(n1,n2))
		dV2 = np.zeros(shape=(n1,n2))
		
		if geometry == None : geometry = 'cartesian'
		
		#------------------------------------------------
		#  define area and volume elements for the
		#  different coordinate systems
		#------------------------------------------------
		
		if geometry == 'cartesian' :
			A1[:] = 1.0
			A2[:] = 1.0
			dV1   = np.outer(dx1,A2)
			dV2   = np.outer(A1,dx2)
		
		if geometry == 'cylindrical' :
			A1 = x1
			A2[:] = 1.0
			dV1 = np.meshgrid(x1*dx1,A2)[0].T*np.meshgrid(x1*dx1,A2)[1].T
			for i in range(n1) : dV2[i,:] = dx2[:]
		
		if geometry == 'polar' :
			A1    = x1
			A2[:] = 1.0
			dV1   = np.meshgrid(x1,A2)[0].T*np.meshgrid(x1,A2)[1].T
			dV2   = np.meshgrid(x1,dx2)[0].T*np.meshgrid(x1,dx2)[1].T

		if geometry == 'spherical' :
			A1 = x1*x1
			A2 = np.sin(x2)
			for j in range(n2): dV1[:,j] = A1*dx1
			dV2   = np.meshgrid(x1,np.sin(x2)*dx2)[0].T*np.meshgrid(x1,np.sin(x2)*dx2)[1].T
		
		# ------------------------------------------------
		#              Make divergence
		# ------------------------------------------------
		for i in range(1,n1-1):
			du1[i,:] = 0.5*(A1[i+1]*u1[i+1,:] - A1[i-1]*u1[i-1,:])/dV1[i,:]
		for j in range(1,n2-1):
			du2[:,j] = 0.5*(A2[j+1]*u2[:,j+1] - A2[j-1]*u2[:,j-1])/dV2[:,j]

		Divergence = du1 + du2
		return Divergence


	def RTh2Cyl(self,R,Th,X1,X2):
		"""
			This method does the transformation from spherical coordinates to cylindrical ones.
			**Inputs:**
			
			R - 2D array of spherical radius coordinates.\n
			Th - 2D array of spherical theta-angle coordinates.\n
			X1 - 2D array of radial component of given vector\n
			X2 - 2D array of thetoidal component of given vector\n
			
			**Outputs:**
			
			This routine outputs two 2D arrays after transformation.
			
			**Usage:**
			
			``import pyPLUTO as pp``\n
			``import numpy as np``\n
			``D = pp.pload(0)``\n
			``ppt=pp.Tools()``\n
			``TH,R=np.meshgrid(D.x2,D.x1)``\n
			``Br,Bz=ppt.RTh2Cyl(R,TH,D.bx1,D.bx2)``
			
			D.bx1 and D.bx2 should be vectors in spherical coordinates. After transformation (Br,Bz) corresponds to vector in cilindrical coordinates.
		"""
		Y1=X1*np.sin(Th)+X2*np.cos(Th)
		Y2=X1*np.cos(Th)-X2*np.sin(Th)
		return Y1,Y2
			
			
	def myInterpol(self,RR,N):
		"""
			This method interpolates (linear interpolation) vector 1D vector RR to 1D N-length vector. Useful for stretched grid calculations.
			**Inputs:**
			RR - 1D array to interpolate.\n
			N  - Number of grids to interpolate to.\n
			**Outputs:**
			This routine outputs interpolated 1D array to the new grid (len=N).
			**Usage:**
			
			``import pyPLUTO as pp``\n
			``import numpy as np``\n
			``D = pp.pload(0)``\n
			``ppt=pp.Tools()``\n
			``x=linspace(0,1,10) #len(x)=10``\n
			``y=x*x``\n
			``Ri,Ni=ppt.myInterpol(y,100) #len(Ri)=100``
			
			Ri - interpolated numbers;
			Ni - grid for Ri
		"""
		
		NN=np.linspace(0,len(RR)-1,len(RR))
		spline_fit=UnivariateSpline(RR,NN,k=3,s=0)
		RRi=np.linspace(RR[0],RR[-1],N)
		NNi=spline_fit(RRi)
		NNi[0]=NN[0]+0.00001
		NNi[-1]=NN[-1]-0.00001
		return RRi,NNi


	def getUniformGrid(self,r,th,rho,Nr,Nth):
		""" This method transforms data with non-uniform grid (stretched) to uniform. Useful for stretched grid calculations.

			**Inputs:**
			r  - 1D vector of X1 coordinate (could be any, e.g D.x1).\n
			th - 1D vector of X2 coordinate (could be any, e.g D.x2).\n
			rho- 2D array of data.\n
			Nr - new size of X1 vector.\n
			Nth- new size of X2 vector.\n
			
			**Outputs:**
			
			This routine outputs 2D uniform array Nr x Nth dimension
			
			**Usage:**
			
			``import pyPLUTO as pp``\n
			``import numpy as np``\n
			``D = pp.pload(0)``\n
			``ppt=pp.Tools()``\n
			``X1new, X2new, res = ppt.getUniformGrid(D.x1,D.x2,D.rho,20,30)``
			
			X1new - X1 interpolated grid len(X1new)=20
			X2new - X2 interpolated grid len(X2new)=30
			res   - 2D array of interpolated variable
			
		"""
		Ri,NRi=self.myInterpol(r,Nr)
		Ra=np.int32(NRi);Wr=NRi-Ra
			
		YY=np.ones([Nr,len(th)])
		for i in range(len(th)):
			YY[:,i]=(1-Wr)*rho[Ra,i] + Wr*rho[Ra+1,i]
					
		THi,NTHi=self.myInterpol(th,Nth)
		THa=np.int32(NTHi);Wth=NTHi-THa
		ZZ=np.ones([Nr,Nth])
		for i in range(Nr):
			ZZ[i,:]=(1-Wth)*YY[i,THa] + Wth*YY[i,THa+1]
		return Ri,THi,ZZ
			
	def congrid(self, a, newdims, method='linear', centre=False, minusone=False):
		"""
		Arbitrary resampling of source array to new dimension sizes.
		Currently only supports maintaining the same number of dimensions.
		To use 1-D arrays, first promote them to shape (x,1).
		
		Uses the same parameters and creates the same co-ordinate lookup points
		as IDL''s congrid routine, which apparently originally came from a VAX/VMS
		routine of the same name.
		
		**Inputs:**
		
		a -- The 2D array that needs resampling into new dimensions.\n
		newdims -- A tuple which represents the shape of the resampled data\n
		method -- This keyword decides the method used for interpolation.\n
		neighbour - closest value from original data\n
		nearest and linear - uses n x 1-D interpolations using scipy.interpolate.interp1d
		(see Numerical Recipes for validity of use of n 1-D interpolations)\n
		spline - uses ndimage.map_coordinates\n
		centre -- This keyword decides the positions of interpolation points.\n
		True - interpolation points are at the centres of the bins\n
		False - points are at the front edge of the bin\n
		minusone -- This prevents extrapolation one element beyond bounds of input array\n
		For example- inarray.shape = (i,j) & new dimensions = (x,y)\n
		False - inarray is resampled by factors of (i/x) * (j/y)\n
		True - inarray is resampled by(i-1)/(x-1) * (j-1)/(y-1)\n
		
		**Outputs:**
		
		A 2D array with resampled data having a shape corresponding to newdims.
	
		"""
		if not a.dtype in [np.float64, np.float32]:
			a = np.cast[float](a)
		m1 = np.cast[int](minusone)
		ofs = np.cast[int](centre) * 0.5
		old = np.array( a.shape )
		ndims = len( a.shape )
		if len( newdims ) != ndims:
			print("[congrid] dimensions error. This routine currently only support rebinning to the same number of dimensions.")
			return None
		newdims = np.asarray( newdims, dtype=float )
		dimlist = []
		
		if method == 'neighbour':
			for i in range( ndims ):
					base = np.indices(newdims)[i]
					dimlist.append((old[i] - m1) / (newdims[i] - m1)*(base + ofs)-ofs)
			cd = np.array( dimlist ).round().astype(int)
			newa = a[list( cd )]
			return newa
		
		elif method in ['nearest','linear']:
				# calculate new dims
			for i in range( ndims ):
					base = np.arange( newdims[i])
					dimlist.append( (old[i] - m1) / (newdims[i] - m1) * (base + ofs) - ofs )
			# specify old dims
			olddims = [np.arange(i, dtype = np.float) for i in list( a.shape )]
			
			# first interpolation - for ndims = any
			mint = interp1d( olddims[-1], a, kind=method )
			newa = mint( dimlist[-1] )
			trorder = [ndims - 1] + list(range( ndims - 1 ))
			for i in range( ndims - 2, -1, -1 ):
					newa = newa.transpose( trorder )
					mint = interp1d( olddims[i], newa, kind=method )
					newa = mint( dimlist[i] )
			
			if ndims > 1:
			# need one more transpose to return to original dimensions
				newa = newa.transpose( trorder )
			return newa
		
		elif method in ['spline']:
			oslices = [ slice(0,j) for j in old ]
			oldcoords = np.ogrid[oslices]
			nslices = [ slice(0,j) for j in list(newdims) ]
			newcoords = np.mgrid[nslices]
			newcoords_dims = range(np.rank(newcoords))
			#make first index last
			newcoords_dims.append(newcoords_dims.pop(0))
			newcoords_tr = newcoords.transpose(newcoords_dims)
			# makes a view that affects newcoords
			newcoords_tr += ofs
			deltas = (np.asarray(old) - m1) / (newdims - m1)
			newcoords_tr *= deltas
			newcoords_tr -= ofs
			newa = scipy.ndimage.map_coordinates(a, newcoords)
			return newa
		else:
			print("Congrid error: Unrecognized interpolation type. Currently only \'neighbour\', \'nearest\',\'linear\' and \'spline\' are supported.")
			return None


	def sph2cyl(self,D,Dx,rphi=None,theta0=None):
		"""
			This method transforms spherical data into cylindrical applying interpolation.
			Works for stretched grid as well, transforms poloidal (R-Theta) data by default.
			Fix theta and set rphi=True to get (R-Phi) transformation.
			
			**Inputs:**
			
			D  - structure  from 'pload' method.\n
			Dx - variable to be transformed (D.rho for example).\n
			
			**Outputs:**
			
			This routine outputs transformed (sph->cyl) variable and grid.
			
			**Usage:**
			
			``import pyPLUTO as pp``\n
			``import numpy as np``\n
			``D = pp.pload(0)``\n
			``ppt=pp.Tools()``\n
			``R,Z,res = ppt.sph2cyl(D,D.rho.transpose())``
			
			R - 2D array with cylindrical radius values
			Z - 2D array with cylindrical Z values
			res - 2D array of transformed variable
			
		"""
		if rphi is None or rphi == False:
			rx=D.x1
			th=D.x2
		else:
			rx=D.x1*np.sin(theta0)
			th=D.x3

		rx,th,Dx=self.getUniformGrid(rx,th,Dx.T,200,200)
		Dx=Dx.T

		if rphi is None or rphi == False:
			r0=np.min(np.sin(th)*rx[0])
			rN=rx[-1]
		else:
			r0=np.min([np.sin(th)*rx[0] , np.sin(th)*rx[-1]])
			rN=np.max([np.sin(th)*rx[0] , np.sin(th)*rx[-1]])

		dr=rN-r0
		z0=np.min(np.cos(th)*rN)
		zN=np.max(np.cos(th)*rN)
		dz=zN-z0
		dth=th[-1]-th[0]
		rl=np.int32(len(rx)*dr/(rx[-1]-rx[0]))
		zl=np.int32(rl* dz/dr)
		thl=len(th)
		r=np.linspace(r0, rN, rl)
		z=np.linspace(z0, zN, zl)
																		
		R,Z = np.meshgrid(r, z)
		Rs = np.sqrt(R*R + Z*Z)
		
		Th = np.arccos(Z/Rs)
		kv_34=self.find(R<0)
		Th.flat[kv_34]=2*np.pi - Th.flat[kv_34]
		ddr=rx[1]-rx[0]
		ddth=th[1]-th[0]

		Rs_copy=Rs.copy()
		Th_copy=Th.copy()

		nR1=self.find(Rs<rx[0])
		Rs.flat[nR1]=rx[0]
		nR2=self.find(Rs>rN)
		Rs.flat[nR2]=rN

		nTh1=self.find(Th>th[-1])
		Th.flat[nTh1]=th[-1]
		nTh2=self.find(Th<th[0])
		Th.flat[nTh2]=th[0]


		ra = ((len(rx)-1.001)/(np.max(Rs.flat)-np.min(Rs.flat)) *(Rs-np.min(Rs.flat)))
		tha = ((thl-1.001)/dth *(Th-th[0]))

		rn = np.int32(ra)
		thn = np.int32(tha)
		dra=ra-rn
		dtha=tha-thn
		w1=1-dra
		w2=dra
		w3=1-dtha
		w4=dtha
		lrx=len(rx)
		NN1=np.int32(rn+thn*lrx)
		NN2=np.int32((rn+1)+thn*lrx)
		NN3=np.int32(rn+(thn+1)*lrx)
		NN4=np.int32((rn+1)+(thn+1)*lrx)
		n=np.transpose(np.arange(0,np.product(np.shape(R))))
		DD=Dx.copy()
		F=R.copy()
		F.flat[n]=w1.flat[n]*(w3.flat[n]*Dx.flat[NN1.flat[n]] + w4.flat[n]*Dx.flat[NN3.flat[n]] )+\
			w2.flat[n]*(w3.flat[n]*Dx.flat[NN2.flat[n]] + w4.flat[n]*Dx.flat[NN4.flat[n]] )
				
		nR1=self.find(Rs_copy<rx[0]-ddr/1.5)
		nR2=self.find(Rs_copy>rN+ddr/1.5)
		nTh1=self.find(Th_copy>th[-1]+ddth/1.5)
		nTh2=self.find(Th_copy<th[0]-ddth/1.5)
			
		nmask=np.concatenate((nR1,nR2,nTh1,nTh2))
		F.flat[nmask]=np.nan;
		return R,Z,F







#File type: extension <Module> set
#By Junxiang H., 2023/07/1
#wacmk.com/cn Tech. Supp.

#This script updates automaticly! Do not Modify!
#Update time:2023-10-24 06:22:13

IO={}
try:
	try:
		import ShockFinder.Addon.IO.HDF5 as HDF5
	except Exception as err:
		print("HDF5in ShockFinder dir importing error, reimporting in current dir")
		import HDF5
	IO["HDF5"]=HDF5
except Exception as err:
	print("Module: HDF5 import failure:",err)

if __name__=="__main__":
	print("Testing Model:",__file__)
	print("IO:")
	for i in IO.keys():
		print(i,":",IO[i])
		print("\t",IO[i].info)


#File type: extension <Module> set
#By Junxiang H., 2023/07/1
#wacmk.com/cn Tech. Supp.

#This script updates automaticly! Do not Modify!
#Update time:2023-10-24 06:22:13

AnalysisTool={}
try:
	try:
		import ShockFinder.Addon.AnalysisTool.SoundSpeed as SoundSpeed
	except Exception as err:
		print("SoundSpeedin ShockFinder dir importing error, reimporting in current dir")
		import SoundSpeed
	AnalysisTool["SoundSpeed"]=SoundSpeed
except Exception as err:
	print("Module: SoundSpeed import failure:",err)

try:
	try:
		import ShockFinder.Addon.AnalysisTool.MachNumber as MachNumber
	except Exception as err:
		print("MachNumberin ShockFinder dir importing error, reimporting in current dir")
		import MachNumber
	AnalysisTool["MachNumber"]=MachNumber
except Exception as err:
	print("Module: MachNumber import failure:",err)

try:
	try:
		import ShockFinder.Addon.AnalysisTool.MassFlux as MassFlux
	except Exception as err:
		print("MassFluxin ShockFinder dir importing error, reimporting in current dir")
		import MassFlux
	AnalysisTool["MassFlux"]=MassFlux
except Exception as err:
	print("Module: MassFlux import failure:",err)

try:
	try:
		import ShockFinder.Addon.AnalysisTool.Shocks as Shocks
	except Exception as err:
		print("Shocksin ShockFinder dir importing error, reimporting in current dir")
		import Shocks
	AnalysisTool["Shocks"]=Shocks
except Exception as err:
	print("Module: Shocks import failure:",err)

try:
	try:
		import ShockFinder.Addon.AnalysisTool.Entropy as Entropy
	except Exception as err:
		print("Entropyin ShockFinder dir importing error, reimporting in current dir")
		import Entropy
	AnalysisTool["Entropy"]=Entropy
except Exception as err:
	print("Module: Entropy import failure:",err)

try:
	try:
		import ShockFinder.Addon.AnalysisTool.Temperature as Temperature
	except Exception as err:
		print("Temperaturein ShockFinder dir importing error, reimporting in current dir")
		import Temperature
	AnalysisTool["Temperature"]=Temperature
except Exception as err:
	print("Module: Temperature import failure:",err)

try:
	try:
		import ShockFinder.Addon.AnalysisTool.Bremsstrahlung as Bremsstrahlung
	except Exception as err:
		print("Bremsstrahlungin ShockFinder dir importing error, reimporting in current dir")
		import Bremsstrahlung
	AnalysisTool["Bremsstrahlung"]=Bremsstrahlung
except Exception as err:
	print("Module: Bremsstrahlung import failure:",err)

if __name__=="__main__":
	print("Testing Model:",__file__)
	print("AnalysisTool:")
	for i in AnalysisTool.keys():
		print(i,":",AnalysisTool[i])
		print("\t",AnalysisTool[i].get)


#File type: algroithm <Function> set
#By Junxiang H., 2023/07/01
#wacmk.com/cn Tech. Supp.

import numpy as np
def get_dx(a):
	if a.ndim==3:
		return a[1:,:,:]-a[:-1,:,:]
	elif a.ndim==2:
		return a[1:,:]-a[:-1,:]
	else:
		return a[1:]-a[:-1]

def get_x(a):
	if a.ndim==3:
		return (a[1:,:,:]+a[:-1,:,:])/2
	elif a.ndim==2:
		return (a[1:,:]+a[:-1,:])/2
	else:
		return (a[1:]+a[:-1])/2

def get_dy(a):
	if a.ndim==3:
		return a[:,1:,:]-a[:,:-1,:]
	elif a.ndim==2:
		return a[:,1:]-a[:,:-1]
	else:
		return None

def get_y(a):
	if a.ndim==3:
		return (a[:,1:,:]+a[:,:-1,:])/2
	elif a.ndim==2:
		return (a[:,1:]+a[:,:-1])/2
	else:
		return None

def get_dz(a):
	if a.ndim==3:
		return a[:,:,1:]-a[:,:,:-1]
	else:
		return None

def get_z(a):
	if a.ndim==3:
		return (a[:,:,1:]+a[:,:,:-1])/2
	else:
		return None

def gradient(Dataobj,quantity_name):
	if type(quantity_name) in (np.ndarray,list,tuple):
		for i in quantity_name:
			Dataobj=gradient(Dataobj,i)
		return Dataobj
	if "geometry" not in Dataobj.quantities.keys() or quantity_name not in Dataobj.quantities.keys():
		print("Warning: args: geometry and",quantity_name,"are needed without definding in 3d gradient calculation")
		return Dataobj
	quantities={}
	if Dataobj.quantities[quantity_name].ndim==1:#1d
		quantities["Gradient_"+quantity_name+"_x1"]=np.gradient(Dataobj.quantities[quantity_name],Dataobj.grid["x1"])
	elif Dataobj.quantities[quantity_name].ndim==2:#2d
		gd=np.gradient(Dataobj.quantities[quantity_name],Dataobj.grid["x1"],Dataobj.grid["x2"])
		if Dataobj.quantities["geometry"] in ("SPHERICAL","POLAR"):
			quantities["Gradient_"+quantity_name+"_x1"]=gd[0]
			quantities["Gradient_"+quantity_name+"_x2"]=gd[1]/Dataobj.grid["x1"].reshape((len(Dataobj.grid["x1"]),1))
		else:
			quantities["Gradient_"+quantity_name+"_x1"]=gd[0]
			quantities["Gradient_"+quantity_name+"_x2"]=gd[1]
	elif Dataobj.quantities[quantity_name].ndim==3:#3d
		gd=np.gradient(Dataobj.quantities[quantity_name],Dataobj.grid["x1"],Dataobj.grid["x2"],Dataobj.grid["x3"])
		if Dataobj.quantities["geometry"] in ("SPHERICAL","POLAR"):
			quantities["Gradient_"+quantity_name+"_x1"]=gd[0]
			quantities["Gradient_"+quantity_name+"_x2"]=gd[1]/Dataobj.grid["x1"].reshape((len(Dataobj.grid["x1"]),1,1))
			if Dataobj.quantities["geometry"]=="SPHERICAL":
				quantities["Gradient_"+quantity_name+"_x3"]=gd[2]/Dataobj.grid["x1"].reshape((len(Dataobj.grid["x1"]),1,1))/np.sin(Dataobj.grid["x2"].reshape((1,len(Dataobj.grid["x2"]),1)))
			else:
				quantities["Gradient_"+quantity_name+"_x3"]=gd[2]
		else:
			quantities["Gradient_"+quantity_name+"_x1"]=gd[0]
			quantities["Gradient_"+quantity_name+"_x2"]=gd[2]
			quantities["Gradient_"+quantity_name+"_x3"]=gd[3]
	Dataobj.update(quantities)
	return Dataobj

def divergence_xyz(F, x=None, y=None, z=None):
	"""
	Compute the divergence of a vector field F in XYZ coordinates.
	
	Parameters:
	- F: Multidimensional array where the last dimension corresponds to the components of the field.
		  F can be 1D ([Fx]), 2D ([Fx, Fy]), or 3D ([Fx, Fy, Fz]).
	- x, y, z: Arrays representing the grids in the x, y, and z directions, respectively
	
	Returns:
	- divF: Array representing the divergence of F. Shape corresponds to the input field.
	"""
	
	if F.ndim == 1:  # 1D
		return np.gradient(F, x)
	
	elif F.ndim == 2:  # 2D
		dFxdx = np.gradient(F[:, 0], x, axis=0)
		dFydy = np.gradient(F[:, 1], y, axis=0)
		return dFxdx + dFydy
	
	elif F.ndim == 3:  # 3D
		dFxdx = np.gradient(F[:, :, 0], x, axis=2)
		dFydy = np.gradient(F[:, :, 1], y, axis=1)
		dFzdz = np.gradient(F[:, :, 2], z, axis=0)
		return dFxdx + dFydy + dFzdz

def divergence_pol(F, r, phi=None, z=None):
	"""
	Compute the divergence of a vector field F in cylindrical coordinates (r, phi, z).
	
	Parameters:
	- F: Multidimensional array where the last dimension corresponds to the components of the field.
		 F can be 1D ([Fr]), 2D ([Fr, Fphi]), or 3D ([Fr, Fphi, Fz]).
	- r, phi, z: Arrays representing the grids in the r, phi, and z directions, respectively.
	
	Returns:
	- divF: Array representing the divergence of F. Shape corresponds to the input field.
	"""
	
	if F.ndim == 1:  # 1D
		return np.gradient(r*F, r)/r
	
	elif F.ndim == 2:  # 2D
		drFr_dr = np.gradient(r*F[:, 0], r, axis=0)
		dFphi_dphi = np.gradient(F[:, 1], phi, axis=1)
		return drFr_dr/r + dFphi_dphi/r
	
	elif F.ndim == 3:  # 3D
		drFr_dr = np.gradient(r*F[:, :, 0], r, axis=2)
		dFphi_dphi = np.gradient(F[:, :, 1], phi, axis=1)
		dFz_dz = np.gradient(F[:, :, 2], z, axis=0)
		return drFr_dr/r + dFphi_dphi/r + dFz_dz

def divergence_sph(F, r, theta=None, phi=None):
	"""
	Compute the divergence of a vector field F in spherical coordinates (r, theta, phi).
	
	Parameters:
	- F: Array where the last dimension corresponds to the components of the field. 
		 F can be 1D (just Fr), 2D ([Fr, Ftheta]) or 3D ([Fr, Ftheta, Fphi]).
	- r, theta, phi: Arrays representing the grids in the r, theta, and phi directions.
	
	Returns:
	- divF: Array representing the divergence of F. Shape corresponds to the input field.
	"""
	
	if F.ndim == 1:  # 1D (just Radial)
		dFr_dr = np.gradient(F, r)
		return (2/r) * F + dFr_dr
	
	elif F.ndim == 2:  # 2D (Radial and Polar angle)
		drFr_dr = np.gradient(r*r*F[:, 0], r, axis=0)
		dFtheta_dtheta = np.gradient(np.sin(theta)*F[:, 1], theta, axis=1)
		return (1/r**2) * drFr_dr + (1/(r * np.sin(theta))) * dFtheta_dtheta
		
	elif F.ndim == 3:  # 3D
		drFr_dr = np.gradient(r*r*F[:, :, 0], r, axis=0)
		dFtheta_dtheta = np.gradient(np.sin(theta)*F[:, :, 1], theta, axis=1)
		dFphi_dphi = np.gradient(F[:, :, 2], phi, axis=2)
		return (1/r**2) * drFr_dr + (1/(r * np.sin(theta))) * (dFtheta_dtheta + dFphi_dphi)

def divergence(Dataobj,quantity_name):
	if type(quantity_name) in (np.ndarray,list,tuple):
		for i in quantity_name:
			Dataobj=divergence(Dataobj,i)
		return Dataobj
	quantities={}
	if "geometry" not in Dataobj.quantities.keys() or quantity_name not in Dataobj.quantities.keys():
			print("Warning: args: geometry and",quantity_name,"are needed without definding in 3d gradient calculation")
			return Dataobj
	if Dataobj.quantities["geometry"]=="SPHERICAL":
		fun=divergence_sph
	elif Dataobj.quantities["geometry"]=="POLAR":
		fun=divergence_pol
	else:
		fun=divergence_xyz
	try:
		quantities["Divergence_"+quantity_name]=fun(Dataobj.quantities[quantity_name],Dataobj.grid["x1"],Dataobj.grid["x2"],Dataobj.grid["x3"])
	except:
		try:
			quantities["Divergence_"+quantity_name]=fun(Dataobj.quantities[quantity_name],Dataobj.grid["x1"],Dataobj.grid["x2"])
		except:
			quantities["Divergence_"+quantity_name]=fun(Dataobj.quantities[quantity_name],Dataobj.grid["x1"])


	if Dataobj.quantities[quantity_name].ndim==1:#1d
		if Dataobj.quantities["geometry"] in ("SPHERICAL","POLAR"):
			if Dataobj.quantities["geometry"]=="SPHERICAL":
				quantities["Divergence_"+quantity_name]=np.gradient(Dataobj.grid["x1"]**2*Dataobj.quantities[quantity_name],Dataobj.grid["x1"])/Dataobj.grid["x1"]**2
			else:
				quantities["Divergence_"+quantity_name]=np.gradient(Dataobj.grid["x1"]*Dataobj.quantities[quantity_name],Dataobj.grid["x1"])/Dataobj.grid["x1"]
		else:
			quantities["Divergence_"+quantity_name]=np.gradient(Dataobj.quantities[quantity_name],Dataobj.grid["x1"])
	elif Dataobj.quantities[quantity_name].ndim==2:#2d
		if Dataobj.quantities["geometry"] in ("SPHERICAL","POLAR"):
			if Dataobj.quantities["geometry"]=="SPHERICAL":
				gd1=np.gradient(Dataobj.grid["x1"].reshape((len(Dataobj.grid["x1"]),1))**2*Dataobj.quantities[quantity_name],Dataobj.grid["x1"],Dataobj.grid["x2"])[0]/Dataobj.grid["x1"].reshape((len(Dataobj.grid["x1"]),1))**2
				gd2=np.gradient(np.sin(Dataobj.grid["x2"].reshape((1,len(Dataobj.grid["x2"]))))*Dataobj.quantities[quantity_name],Dataobj.grid["x1"],Dataobj.grid["x2"])[1]/Dataobj.grid["x1"].reshape((len(Dataobj.grid["x1"]),1))/np.sin(Dataobj.grid["x2"].reshape((1,len(Dataobj.grid["x2"]))))
			else:
				gd1=np.gradient(Dataobj.grid["x1"].reshape((len(Dataobj.grid["x1"]),1))*Dataobj.quantities[quantity_name],Dataobj.grid["x1"],Dataobj.grid["x2"])[0]/Dataobj.grid["x1"].reshape((len(Dataobj.grid["x1"]),1))
				gd2=np.gradient(Dataobj.quantities[quantity_name],Dataobj.grid["x1"],Dataobj.grid["x2"])[1]/Dataobj.grid["x1"].reshape((len(Dataobj.grid["x1"]),1))
		else:
			gd1,gd2=np.gradient(Dataobj.quantities[quantity_name],Dataobj.grid["x1"],Dataobj.grid["x2"])
		quantities["Divergence_"+quantity_name]=gd1+gd2
	elif Dataobj.quantities[quantity_name].ndim==3:#3d
		if Dataobj.quantities["geometry"] in ("SPHERICAL","POLAR"):
			if Dataobj.quantities["geometry"]=="SPHERICAL":
				gd1=np.gradient(Dataobj.grid["x1"].reshape((len(Dataobj.grid["x1"]),1,1))**2*Dataobj.quantities[quantity_name],Dataobj.grid["x1"],Dataobj.grid["x2"],Dataobj.grid["x3"])[0]/Dataobj.grid["x1"].reshape((len(Dataobj.grid["x1"]),1,1))**2
				gd2=np.gradient(np.sin(Dataobj.grid["x2"].reshape((1,len(Dataobj.grid["x2"]),1)))*Dataobj.quantities[quantity_name],Dataobj.grid["x1"],Dataobj.grid["x2"],Dataobj.grid["x3"])[1]/Dataobj.grid["x1"].reshape((len(Dataobj.grid["x1"]),1,1))/np.sin(Dataobj.grid["x2"].reshape((1,len(Dataobj.grid["x2"]),1)))
				gd3=np.gradient(Dataobj.quantities[quantity_name],Dataobj.grid["x1"],Dataobj.grid["x2"],Dataobj.grid["x3"])[2]/Dataobj.grid["x1"].reshape((len(Dataobj.grid["x1"]),1,1))/np.sin(Dataobj.grid["x2"].reshape((1,len(Dataobj.grid["x2"]),1)))
			else:
				gd1=np.gradient(Dataobj.grid["x1"].reshape((len(Dataobj.grid["x1"]),1,1))*Dataobj.quantities[quantity_name],Dataobj.grid["x1"],Dataobj.grid["x2"],Dataobj.grid["x3"])[0]/Dataobj.grid["x1"].reshape((len(Dataobj.grid["x1"]),1,1))
				gd2=np.gradient(Dataobj.quantities[quantity_name],Dataobj.grid["x1"],Dataobj.grid["x2"],Dataobj.grid["x3"])[1]/Dataobj.grid["x1"].reshape((len(Dataobj.grid["x1"]),1,1))
				gd3=np.gradient(Dataobj.quantities[quantity_name],Dataobj.grid["x1"],Dataobj.grid["x2"],Dataobj.grid["x3"])[2]
		else:
			gd1,gd2,gd3=np.gradient(Dataobj.quantities[quantity_name],Dataobj.grid["x1"],Dataobj.grid["x2"],Dataobj.grid["x3"])
		quantities["Divergence_"+quantity_name]=gd1+gd2+gd3
	Dataobj.update(quantities)
	return Dataobj

def result(quantity_name=None,anafname=None):
	if anafname=="Gradient":
		return ["Gradient_"+i+"_x1" for i in quantity_name.split(",")]#+["Gradient_"+i+"_x2" for i in quantity_name.split(",")]+["Gradient_"+i+"_x3" for i in quantity_name.split(",")]
	elif anafname=="Divergence":
		return ["Divergence_"+i for i in quantity_name.split(",")]
	return ()

'''
#abort!
def integrate(fx,x): #inte@ f(x)dx
	if len(x)>0:
		result=integrate(fx,x[1:])
		if type(result) !=np.ndarray or result.ndim==0:
			result=np.full((len(x[0]),),result)
		return np.trapz(result,x=x[0])	
	return fx

def integrate_single(fx,x):
	if type(fx) !=np.ndarray or fx.ndim==0:
		fx=np.full((len(x),),fx)
	return np.trapz(fx,x)


def integrate_body(f,x,y=None,z=None):
	if type(f)!=np.ndarray or f.ndim<=1:
		if type(y)==None:
			y=np.arange(0,1,1/len(x))
		if type(z)==None:
			z=np.arange(0,1,1/len(x))
		return integrate_single(integrate(f,(y,z)),x)
	elif f.ndim==2:
		if type(y)==None:
			y=np.arange(0,1,1/len(f[0]))
		if type(z)==None:
			z=np.arange(0,1,1/len(f[0]))
		return integrate(integrate_single(f,z),(x,y))
	elif f.ndim==3:
		if type(y)==None:
			y=np.arange(0,1,1/len(f[0]))
		if type(z)==None:
			z=np.arange(0,1,1/len(f[0][0]))
		return integrate(f,(x,y,z))

def integrate_surface_xoy(f,x=None,y=None,default_step=10000):
	if type(f)!=np.ndarray or f.ndim<=1:
		if type(x)==None:
			x=np.arange(0,1,1/len(f))
		if type(y)==None:
			y=np.arange(0,1,1/default_step)
		return integrate(f,(x,y))
	elif f.ndim>=2:
		if type(x)==None:
			x=np.arange(0,1,1/len(f))
		if type(y)==None:
			y=np.arange(0,1,1/len(f[0]))
		return integrate(f,(x,y))

def integrate_sph_body(f,r,th=None,phi=None):
	if type(f) !=np.ndarray or f.ndim<=1:
		if type(th)==type(None):
			th=np.arange(0,np.pi,np.pi/len(r))
		if type(phi)==type(None):
			phi=np.arange(0,2*np.pi,2*np.pi/len(r))
		return integrate_single(f*r**2*integrate(np.sin(th),(th,phi)),r)
	elif f.ndim==2:#r th
		if type(th)==type(None):
			th=np.arange(0,np.pi,np.pi/len(f[0]))
		if type(phi)==type(None):
			phi=np.arange(0,2*np.pi,2*np.pi/len(r))
		return integrate(f*r.reshape((len(r),1))**2*np.sin(th.reshape((1,len(th))))*integrate_single(1,phi),(r,th))
	else: #r th phi
		if type(th)==type(None):
			th=np.arange(0,np.pi,np.pi/len(f[0]))
		if type(phi)==type(None):
			phi=np.arange(0,2*np.pi,2*np.pi/len(f[0][0]))
		return integrate(f*r.reshape((len(r),1,1))**2*np.sin(th.reshape((1,len(th),1))),(r,th,phi))

def integrate_sph_sur(f,r,th=None,phi=None,default_step=10000): #r is fixed
	if type(th)==type(None):
		if type(f) !=np.ndarray or f.ndim==0:
			th=np.arange(0,np.pi,np.pi/default_step)
		elif f.ndim==1:
			th=np.arange(0,np.pi,np.pi/len(f))
		elif f.ndim==2:
			th=np.arange(0,np.pi,np.pi/len(f[0]))
		elif f.ndim==3:
			th=np.arange(0,np.pi,np.pi/len(f[0][0]))
	if type(phi)==type(None):
		if type(f) !=np.ndarray or f.ndim==0:
			phi=np.arange(0,2*np.pi,2*np.pi/default_step)
		elif f.ndim==1:
			phi=np.arange(0,2*np.pi,2*np.pi/len(f))
		elif f.ndim==2:
			phi=np.arange(0,2*np.pi,2*np.pi/len(f[0]))
		elif f.ndim==3:
			phi=np.arange(0,2*np.pi,2*np.pi/len(f[0][0]))
	if type(f) !=np.ndarray or f.ndim<=1: #th
		return integrate_single(f*r**2*np.sin(th)*integrate_single(1,phi),th)
	elif f.ndim==2: #th phi
		return integrate(f*r**2*np.sin(th.reshape((1,len(th)))),(th,phi))
	elif f.ndim==3: #th phi
		return integrate(f*r**2*np.sin(th.reshape((1,len(th),1))),(th,phi))

def integrate_pol_body(f,r,phi=None,z=None):
	#f=f(x,th,z)
	if type(f) !=np.ndarray or f.ndim<=1: #r 
		if type(phi)==type(None):
			phi=np.arange(0,2*np.pi,2*np.pi/len(r))
		if type(z)==type(None):
			z=np.arange(0,1,1/len(r))
		return integrate_single(f*r*integrate(1,(phi,z)),r)
	elif f.ndim==2:#r phi
		if type(phi)==type(None):
			phi=np.arange(0,2*np.pi,2*np.pi/len(f[0]))
		if type(z)==type(None):
			z=np.arange(0,1,1/len(r))
		return integrate(f*r.reshape((len(r),1))*integrate_single(1,z),(r,phi))
	else: #x z th
		if type(phi)==type(None):
			phi=np.arange(0,2*np.pi,2*np.pi/len(f[0]))
		if type(z)==type(None):
			z=np.arange(0,1,1/len(f[0][0]))
		return integrate(f*r.reshape((len(r),1,1))*phi.reshape((1,len(phi),1)),(r,phi,z))


def integrate_pol_sur_xoz(f,x,z,phi=None,default_step=10000): #x is fixed
	if type(phi)==type(None):
		if type(f) !=np.ndarray or f.ndim==0:
			phi=np.arange(0,np.pi,2*np.pi/default_step)
		elif f.ndim==1:
			phi=np.arange(0,np.pi,2*np.pi/len(f))
		elif f.ndim==2:
			phi=np.arange(0,np.pi,2*np.pi/len(f[0]))
		elif f.ndim==3:
			phi=np.arange(0,np.pi,2*np.pi/len(f[0][0]))
	if type(f) !=np.ndarray or f.ndim<=1: #z
		return integrate_single(f*x*integrate_single(1,phi),z)
	elif f.ndim==2: #z th
		return integrate(f*x*phi.reshape((1,len(phi))),(phi,z))
	elif f.ndim==3: #z th
		return integrate(f*x*phi.reshape((1,len(phi),1)),(phi,z))


def integrate_pol_sur_xop(f,x,z,phi=None,default_step=10000): #z is fixed
	if type(phi)==type(None):
		if type(f) !=np.ndarray or f.ndim==0:
			phi=np.arange(0,np.pi,2*np.pi/default_step)
		elif f.ndim==1:
			phi=np.arange(0,np.pi,2*np.pi/len(f))
		elif f.ndim==2:
			phi=np.arange(0,np.pi,2*np.pi/len(f[0]))
		elif f.ndim==3:
			phi=np.arange(0,np.pi,2*np.pi/len(f[0][0]))
	if type(f) !=np.ndarray or f.ndim<=1: #x
		return integrate_single(f*x*integrate_single(1,phi),x)
	elif f.ndim==2: #x th
		return integrate(f*x.reshape((len(x),1))*phi.reshape((1,len(phi))),(x,phi))
	elif f.ndim==3: #x th
		return integrate(f*x.reshape((len(x),1,1))*phi.reshape((1,len(phi),1)),(x,phi))

'''

def integrate_body(fx, x=None, y=None, z=None): #by chat
	if type(f) in (float,int):
		mark=1
		if x!=None:
			mark*=x[-1]-x[0]
		if y!=None:
			mark*=y[-1]-y[0]
		if z!=None:
			mark*=z[-1]-z[0]
		return mark*fx
	if fx.ndim == 1:
		x = np.linspace(0, 1, len(fx)) if x is None else x
		if len(x) != len(fx):
			raise ValueError("Length of x must match the length of 1D fx.")
		result = np.trapz(fx, x)

	elif fx.ndim == 2:
		x = np.linspace(0, 1, fx.shape[0]) if x is None else x
		y = np.linspace(0, 1, fx.shape[1]) if y is None else y

		if len(x) != fx.shape[0]:
			raise ValueError("Length of x must match the first dimension of 2D fx.")
		if len(y) != fx.shape[1]:
			raise ValueError("Length of y must match the second dimension of 2D fx.")

		result = np.trapz(np.trapz(fx, y, axis=-1), x, axis=0)

	elif fx.ndim == 3:
		x = np.linspace(0, 1, fx.shape[0]) if x is None else x
		y = np.linspace(0, 1, fx.shape[1]) if y is None else y
		z = np.linspace(0, 1, fx.shape[2]) if z is None else z

		if len(x) != fx.shape[0]:
			raise ValueError("Length of x must match the first dimension of 3D fx.")
		if len(y) != fx.shape[1]:
			raise ValueError("Length of y must match the second dimension of 3D fx.")
		if len(z) != fx.shape[2]:
			raise ValueError("Length of z must match the third dimension of 3D fx.")

		result = np.trapz(np.trapz(np.trapz(fx, z, axis=-1), y, axis=-1), x, axis=0)

	else:
		raise ValueError("The fx array should be either 1D, 2D, or 3D.")
	
	if not isinstance(result, (float, np.float64)):
		raise ValueError("Integration did not result in a scalar value as expected.")

	return result

def integrate_pol_body2(f, r=None, phi=None, z=None):
	# Default values
	# Default values
	r = np.linspace(0, 1, f.shape[0]) if r is None else r
	phi = np.linspace(0, 2*np.pi, f.shape[1]) if phi is None and f.ndim > 1 else phi
	z = np.linspace(0, 1, f.shape[2]) if z is None and f.ndim == 3 else z

	# Check shapes
	if len(r) != f.shape[0]:
		raise ValueError("Length of r must match the first dimension of f.")
	if f.ndim > 1 and len(phi) != f.shape[1]:
		raise ValueError("Length of phi must match the second dimension of f.")
	if f.ndim == 3 and len(z) != f.shape[2]:
		raise ValueError("Length of z must match the third dimension of f.")

	# Integration
	result = f

	# For 1D data
	if f.ndim == 1:
		result = np.trapz(result * r, r)

	# For 2D and 3D data
	else:
		if f.ndim >= 2:
			result = np.trapz(result * r[:, np.newaxis], r, axis=0)  # account for the r in the differential
			if result.ndim > 1:  # check if the result is still multidimensional
				result = np.trapz(result, phi, axis=-1)
	
		if result.ndim > 1:  # check again, to integrate over z if needed
			result = np.trapz(result, z, axis=-1)

	return result


def integrate_pol_body(f, r=None, phi=None, z=None):
	if type(f) in (float,int):
		mark=1
		if r!=None:
			mark*=0.5*(r[-1]**2-r[0]**2)
		if phi!=None:
			mark*=phi
		else:
			mark*=np.pi
		if z!=None:
			mark*=z[-1]-z[0]
		return mark*f
	# Default values
	r_values = np.linspace(0, 1, f.shape[0]) if r is None else r
	try:
		phi_values = np.linspace(0, 2*np.pi, f.shape[1]) if phi is None else phi
	except:
		phi_values = np.linspace(0, 2*np.pi, 100) if phi is None else phi
	try:
		z_values = np.linspace(0, 1, f.shape[2]) if z is None else z
	except:
		z_values = np.linspace(0, 1, 100) if z is None else z
	# Integration
	result = f

	# 1D data: f is symmetric in both phi and z directions
	if f.ndim == 1:
		result = np.trapz(result * r_values, r_values)
		result *= phi_values[-1] - phi_values[0]  # Integration over phi
		result *= z_values[-1] - z_values[0]  # Integration over z

	# 2D data: f is considered homogeneous in r and z if r isn't provided
	elif f.ndim == 2:
		if r is None:
			result = np.mean(result, axis=0)  # Mean over r (homogeneous)
			result *= 0.5*(r_values[-1]**2 - r_values[0]**2)  # Integration over r
		else:
			result = np.trapz(result*r_values, r_values, axis=0)
		
		result = np.trapz(result, phi_values)
		result *= (z_values[-1] - z_values[0])  # Integration over z

	# 3D data: No assumed symmetries
	else:
		result = np.trapz(result*r_values, r_values, axis=0)
		result = np.trapz(result, phi_values, axis=0)
		result = np.trapz(result, z_values)

	if not isinstance(result, (float, np.float64)):
		raise ValueError("Integration did not result in a scalar value as expected.")

	return result


def integrate_sph_body(f, r=None, th=None, phi=None):
	if type(f) in (float,int):
		mark=1
		if r!=None:
			mark*=(r[-1]**3-r[0]**3)/3
		else:
			mark/=3
		if th!=None:
			mark*=np.cos(th[0])-np.cos(th[-1])
		else:
			mark*=2
		if phi!=None:
			mark*=phi[-1]-phi[0]
		else:
			mark*=2*np.pi
		return mark*f
	# Default values
	r_values = np.linspace(0, 1, f.shape[0]) if r is None else r
	try:
		th_values = np.linspace(0, np.pi, f.shape[1]) if th is None else th
	except:
		th_values = np.linspace(0, np.pi, 100) if th is None else th
	try:
		phi_values = np.linspace(0, 2*np.pi, f.shape[2]) if phi is None else phi
	except:
		phi_values = np.linspace(0, 2*np.pi, 100) if phi is None else phi

	# Integration
	result = f

	# 1D data: f is symmetric in both theta and phi directions
	if f.ndim == 1:
		result = np.trapz(result * r_values**2, r_values)
		result *= phi_values[-1]-phi_values[0]  # Integration over phi
		result *= np.cos(th_values[0])-np.cos(th_values[-1])  # Integration over theta

	# 2D data: f is considered homogeneous in r and phi if r isn't provided
	elif f.ndim == 2:
		if r is None:
			result = np.mean(result, axis=0)  # Mean over r (homogeneous)
			result *= (r_values[-1]**3 - r_values[0]**3)/3  # Integration over r
		else:
			result = np.trapz(result * r_values[:, np.newaxis]**2, r_values, axis=0)
		
		result = np.trapz(result * np.sin(th_values), th_values)
		result *= phi_values[-1]-phi_values[0]  # Integration over phi

	# 3D data: No assumed symmetries
	else:
		result = np.trapz(result * r_values[:, np.newaxis, np.newaxis]**2, r_values, axis=0)
		result = np.trapz(result * np.sin(th_values[:, np.newaxis]), th_values, axis=0)
		result = np.trapz(result, phi_values)

	if not isinstance(result, (float, np.float64)):
		raise ValueError("Integration did not result in a scalar value as expected.")

	return result

def get_closest_index(value, dim):
	return min(range(len(dim)), key=lambda i: abs(dim[i] - value))

def get_indices_for_range(value_range, dim):
	return [get_closest_index(value_range[0], dim), get_closest_index(value_range[1], dim)]

def get_indices_for_ranges(value_ranges, dim):
	return [get_indices_for_range(value_range, dim) for value_range in value_ranges]

def integrate_surface(f, x, y, z=None, xr=(), yr=(), zr=(), surface=()):
	if not surface:
		raise ValueError("Please specify a surface to integrate over.")
		
	if len(xr) > len(surface):
		raise ValueError("Length of xr must not exceed the length of surface.")
	if len(yr) > len(surface):
		raise ValueError("Length of yr must not exceed the length of surface.")
	if type(z)!=type(None) and len(zr) > len(surface):
		raise ValueError("Length of zr must not exceed the length of surface.")

	# Helper function to get index closest to the value

	# Initialize the total flux
	
	result=[]
	for i, surf in enumerate(surface):
		total_flux = 0
		sign = 1 if not surf.startswith("-") else -1
		surf_dim = surf[-1]  # Extract the dimension

		boundaries = {
			"min": 0,
			"max": -1
		}
		
		# Determine range or boundary based on the surface dimension
		if surf_dim == "x":
			boundary = xr[i] if i < len(xr) else "both"
			y_vals = get_indices_for_ranges(yr[i],y) if i < len(yr) and len(yr[i]) > 0 else []
			z_vals = get_indices_for_ranges(zr[i],z) if type(z)!=type(None)  and i < len(zr) and len(zr[i]) > 0 else []
			if y_vals==z_vals and y_vals==[]:
				y_vals.append(range(len(y)))
				if f.ndim==3:
					for j in y_vals:
						z_vals.append(range(len(z)))
				elif len(zr)>i and len(zr[i])>0:
					for j in range(len(y_vals)):
						z_vals.append(zr[i][j])
			elif y_vals==[]:
				for j in z_vals:
					y_vals.append(range(len(y)))
			elif z_vals==[]:
				if f.ndim==3:
					for j in y_vals:
						z_vals.append(range(len(z)))
				elif len(zr)>i and len(zr[i])>0:
					for j in range(len(y_vals)):
						z_vals.append(zr[i][j])

		elif surf_dim == "y":
			boundary = yr[i] if i < len(yr) else "both"
			x_vals = get_indices_for_ranges(xr[i],x) if i < len(xr) and len(xr[i]) > 0 else []
			z_vals = get_indices_for_ranges(zr[i],z) if type(z)!=type(None)  and i < len(zr) and len(zr[i]) > 0 else []
			if x_vals==z_vals and x_vals==[]:
				x_vals.append(range(len(x)))
				if f.ndim==3:
					for j in x_vals:
						z_vals.append(range(len(z)))
				elif len(zr)>i and len(zr[i])>0:
					for j in range(len(x_vals)):
						z_vals.append(zr[i][j])
			elif x_vals==[]:
				for j in z_vals:
					x_vals.append(range(len(x)))
			elif z_vals==[]:
				if f.ndim==3:
					for j in x_vals:
						z_vals.append(range(len(z)))
				elif len(zr)>i and len(zr[i])>0:
					for j in range(len(x_vals)):
						z_vals.append(zr[i][j])

		elif surf_dim == "z":
			boundary = zr[i] if i < len(zr) else "both"
			x_vals = get_indices_for_ranges(xr[i],x) if i < len(xr) and len(xr[i]) > 0 else []
			y_vals = get_indices_for_ranges(yr[i],y) if i < len(yr) and len(yr[i]) > 0 else []
			if x_vals==y_vals and x_vals==[]:
				x_vals.append(range(len(x)))
				y_vals.append(range(len(y)))
			elif x_vals==[]:
				for j in y_vals:
					x_vals.append(range(len(x)))
			elif y_vals==[]:
				for j in x_vals:
					y_vals.append(range(len(y)))

		# Check for 0 values
		if boundary == 0:
			boundary = "min"

		# Calculate flux based on explicit grid values
		if surf_dim == "x":
			for boundary_val in [boundaries[boundary]] if boundary != "both" else [0, -1]:
				for j in range(len(y_vals)):
					slice_data = f[boundary_val, y_vals[j]] if f.ndim == 2 else f[boundary_val, y_vals[j]][:,z_vals[j]]
					tmpl=np.trapz(slice_data, y[y_vals[j]],axis=0)
					if f.ndim==3:
						tmpl=np.trapz(tmpl, z[z_vals[j]])
					elif len(z_vals)>0:
						tmpl*=z_vals[j][-1]-z_vals[j][0]
					total_flux += sign * tmpl

		elif surf_dim == "y":
			for boundary_val in [boundaries[boundary]] if boundary != "both" else [0, -1]:
				for j in range(len(x_vals)):
					slice_data = f[x_vals[j], boundary_val] if f.ndim == 2 else f[x_vals[j], boundary_val][:,z_vals[j]]
					tmpl=np.trapz(slice_data, x[x_vals[j]],axis=0)
					if f.ndim==3:
						tmpl=np.trapz(tmpl, z[z_vals[j]])
					elif len(z_vals)>0:
						tmpl*=z_vals[j][-1]-z_vals[j][0]
					total_flux += sign * tmpl

		elif surf_dim == "z":
			for boundary_val in [boundaries[boundary]] if boundary != "both" else [0, -1]:
				for j in range(len(x_vals)):
					slice_data = f[x_vals[j], : , boundary_val][:,y_vals[j]]
					tmpl=np.trapz(slice_data, x[x_vals[j]],axis=0)
					if f.ndim==3:
						tmpl=np.trapz(tmpl, y[y_vals[j]])
					total_flux += sign * tmpl

		result.append(total_flux)

	return result


def integrate_pol_sur(f, r, phi, z=None, rr=(), pr=(), zr=(), surface=()):
	if not surface:
		raise ValueError("Please specify a surface to integrate over.")
		
	if len(rr) > len(surface):
		raise ValueError("Length of rr must not exceed the length of surface.")
	if len(pr) > len(surface):
		raise ValueError("Length of pr must not exceed the length of surface.")
	if type(z)!=None and len(zr) > len(surface):
		raise ValueError("Length of zr must not exceed the length of surface.")


	result=[]

	for i, surf in enumerate(surface):
		total_flux=0
		sign = 1 if not surf.startswith("-") else -1
		surf_dim = surf.replace("-","")

		boundaries = {
			"min": 0,
			"max": -1
		}

		# Similar to your Cartesian version but adapted for polar coordinates
		if surf_dim == "r":
			boundary = rr[i] if i < len(rr) else "both"
			phi_vals = get_indices_for_ranges(pr[i],phi) if i < len(pr) and len(pr[i]) > 0 else []
			z_vals = get_indices_for_ranges(zr[i],z) if type(z)!=None and i < len(zr) and len(zr[i]) > 0 else []
			if phi_vals==z_vals and phi_vals==[]:
				phi_vals.append(range(len(phi)))
				if f.ndim==3:
					for j in phi_vals:
						z_vals.append(range(len(z)))
				elif len(zr)>i and len(zr[i])>0:
					for j in range(len(phi_vals)):
						z_vals.append(zr[i][j])
			elif phi_vals==[]:
				for j in z_vals:
					phi_vals.append(range(len(phi)))
			elif z_vals==[]:
				if f.ndim==3:
					for j in phi_vals:
						z_vals.append(range(len(z)))
				elif len(zr)>i and len(zr[i])>0:
					for j in range(len(phi_vals)):
						z_vals.append(zr[i][j])

			# ... Rest of the logic ...

		elif surf_dim == "phi":
			boundary = pr[i] if i < len(pr) else "both"
			r_vals = get_indices_for_ranges(rr[i],r) if i < len(rr) and len(rr[i]) > 0 else []
			z_vals = get_indices_for_ranges(zr[i],z) if type(z)!=None and i < len(zr) and len(zr[i]) > 0 else []
			if r_vals==z_vals and r_vals==[]:
				r_vals.append(range(len(r)))
				if f.ndim==3:
					for j in r_vals:
						z_vals.append(range(len(z)))
				elif len(zr)>i and len(zr[i])>0:
					for j in range(len(r_vals)):
						z_vals.append(zr[i][j])
			elif r_vals==[]:
				for j in z_vals:
					r_vals.append(range(len(r)))
			elif z_vals==[]:
				if f.ndim==3:
					for j in r_vals:
						z_vals.append(range(len(z)))
				elif len(zr)>i and len(zr[i])>0:
					for j in range(len(r_vals)):
						z_vals.append(zr[i][j])
			# ... Rest of the logic ...

		elif surf_dim == "z":
			boundary = zr[i] if i < len(zr) else "both"
			r_vals = get_indices_for_ranges(rr[i],r) if i < len(rr) and len(rr[i]) > 0 else []
			phi_vals = get_indices_for_ranges(pr[i],phi) if i < len(pr) and len(pr[i]) > 0 else []
			if r_vals==phi_vals and r_vals==[]:
				r_vals.append(range(len(r)))
				phi_vals.append(range(len(phi)))
			elif r_vals==[]:
				for j in phi_vals:
					r_vals.append(range(len(r)))
			elif phi_vals==[]:
				for j in r_vals:
					phi_vals.append(range(len(phi)))
			# ... Rest of the logic ...

		if boundary == 0:
			boundary = "min"
		# Calculate flux based on explicit grid values
		if surf_dim == "r":
			for boundary_val in [boundaries[boundary]] if boundary != "both" else [0, -1]:
				for j in range(len(phi_vals)):
					slice_data = f[boundary_val, phi_vals[j]] if f.ndim == 2 else f[boundary_val, phi_vals[j]][:,z_vals[j]]
					tmpl=np.trapz(slice_data*r[boundary_val], phi[phi_vals[j]],axis=0)
					if f.ndim==3:
						tmpl=np.trapz(tmpl, z[z_vals[j]])
					elif len(z_vals)>0:
						tmpl*=z_vals[j][-1]-z_vals[j][0]
					total_flux += sign * tmpl

		elif surf_dim == "phi":
			for boundary_val in [boundaries[boundary]] if boundary != "both" else [0, -1]:
				for j in range(len(r_vals)):
					slice_data = f[r_vals[j], boundary_val] if f.ndim == 2 else f[r_vals[j], boundary_val][:,z_vals[j]]
					tmpl=np.trapz(slice_data, r[r_vals[j]],axis=0)
					if f.ndim==3:
						tmpl=np.trapz(tmpl, z[z_vals[j]])
					elif len(z_vals)>0:
						tmpl*=z_vals[j][-1]-z_vals[j][0]
					total_flux += sign * tmpl

		elif surf_dim == "z":
			for boundary_val in [boundaries[boundary]] if boundary != "both" else [0, -1]:
				for j in range(len(r_vals)):
					slice_data = f[r_vals[j], : , boundary_val][:,phi_vals[j]]
					tmpl=np.trapz(slice_data*r[r_vals[j]], r[r_vals[j]],axis=0)
					if f.ndim==3:
						tmpl=np.trapz(tmpl, phi[phi_vals[j]])
					total_flux += sign * tmpl

		result.append(total_flux)
	return result

def integrate_sph_sur(f, r, th, phi=None, rr=(), tr=(), pr=(), surface=()):
	if not surface:
		raise ValueError("Please specify a surface to integrate over.")
		
	if len(rr) > len(surface):
		raise ValueError("Length of rr must not exceed the length of surface.")
	if len(tr) > len(surface):
		raise ValueError("Length of tr must not exceed the length of surface.")
	if type(phi)!=None and len(pr) > len(surface):
		raise ValueError("Length of pr must not exceed the length of surface.")


	result=[]

	for i, surf in enumerate(surface):
		total_flux=0
		sign = 1 if not surf.startswith("-") else -1
		surf_dim = surf.replace("-","")

		boundaries = {
			"min": 0,
			"max": -1
		}

		# Similar to your Cartesian version but adapted for polar coordinates
		if surf_dim == "r":
			boundary = rr[i] if i < len(rr) else "both"
			th_vals = get_indices_for_ranges(tr[i],th) if i < len(tr) and len(tr[i]) > 0 else []
			phi_vals = get_indices_for_ranges(pr[i],phi) if type(phi)!=None and i < len(pr) and len(pr[i]) > 0 else []
			if th_vals==phi_vals and th_vals==[]:
				th_vals.append(range(len(th)))
				if f.ndim==3:
					for j in th_vals:
						phi_vals.append(range(len(phi)))
				elif len(pr)>i and len(pr[i])>0:
					for j in range(len(th_vals)):
						phi_vals.append(pr[i][j])
			elif th_vals==[]:
				for j in z_vals:
					th_vals.append(range(len(th)))
			elif phi_vals==[]:
				if f.ndim==3:
					for j in th_vals:
						phi_vals.append(range(len(z)))
				elif len(pr)>i and len(pr[i])>0:
					for j in range(len(th_vals)):
						phi_vals.append(pr[i][j])

			# ... Rest of the logic ...

		elif surf_dim == "th":
			boundary = tr[i] if i < len(tr) else "both"
			r_vals = get_indices_for_ranges(rr[i],r) if i < len(rr) and len(rr[i]) > 0 else []
			phi_vals = get_indices_for_ranges(pr[i],phi) if type(phi)!=None and i < len(pr) and len(pr[i]) > 0 else []
			if r_vals==phi_vals and r_vals==[]:
				r_vals.append(range(len(r)))
				if f.ndim==3:
					for j in r_vals:
						phi_vals.append(range(len(phi)))
				elif len(pr)>i and len(pr[i])>0:
					for j in range(len(r_vals)):
						phi_vals.append(pr[i][j])
			elif r_vals==[]:
				for j in phi_vals:
					r_vals.append(range(len(r)))
			elif phi_vals==[]:
				if f.ndim==3:
					for j in r_vals:
						phi_vals.append(range(len(phi)))
				elif len(pr)>i and len(pr[i])>0:
					for j in range(len(r_vals)):
						phi_vals.append(pr[i][j])
			# ... Rest of the logic ...

		elif surf_dim == "phi":
			boundary = pr[i] if i < len(pr) else "both"
			r_vals = get_indices_for_ranges(rr[i],r) if i < len(rr) and len(rr[i]) > 0 else []
			th_vals = get_indices_for_ranges(tr[i],th) if i < len(tr) and len(tr[i]) > 0 else []
			if r_vals==th_vals and r_vals==[]:
				r_vals.append(range(len(r)))
				th_vals.append(range(len(th)))
			elif r_vals==[]:
				for j in th_vals:
					r_vals.append(range(len(r)))
			elif th_vals==[]:
				for j in r_vals:
					th_vals.append(range(len(th)))
			# ... Rest of the logic ...

		if boundary == 0:
			boundary = "min"
		# Calculate flux based on explicit grid values
		if surf_dim == "r":
			for boundary_val in [boundaries[boundary]] if boundary != "both" else [0, -1]:
				for j in range(len(th_vals)):
					slice_data = f[boundary_val, th_vals[j]] if f.ndim == 2 else f[boundary_val, th_vals[j]][:,phi_vals[j]]
					tmpl=np.trapz(slice_data*r[boundary_val]**2*np.sin(th[th_vals[j]]), th[th_vals[j]],axis=0)
					if f.ndim==3:
						tmpl=np.trapz(tmpl, phi[phi_vals[j]])
					elif len(phi_vals)>0:
						tmpl*=phi_vals[j][-1]-phi_vals[j][0]
					else:
						tmpl*=2*np.pi
					total_flux += sign * tmpl

		elif surf_dim == "th":
			for boundary_val in [boundaries[boundary]] if boundary != "both" else [0, -1]:
				for j in range(len(r_vals)):
					slice_data = f[r_vals[j], boundary_val] if f.ndim == 2 else f[r_vals[j], boundary_val][:,phi_vals[j]]
					tmpl=np.trapz(slice_data*np.sin(th[boundary_val])*r[r_vals[j]], r[r_vals[j]],axis=0)
					if f.ndim==3:
						tmpl=np.trapz(tmpl, phi[phi_vals[j]])
					elif len(phi_vals)>0:
						tmpl*=phi_vals[j][-1]-phi_vals[j][0]
					else:
						tmpl*=2*np.pi
					total_flux += sign * tmpl

		elif surf_dim == "phi":
			for boundary_val in [boundaries[boundary]] if boundary != "both" else [0, -1]:
				for j in range(len(r_vals)):
					slice_data = f[r_vals[j], : , boundary_val][:,th_vals[j]]
					tmpl=np.trapz(slice_data*r[r_vals[j]], r[r_vals[j]],axis=0)
					if f.ndim==3:
						tmpl=np.trapz(tmpl, th[th_vals[j]])
					total_flux += sign * tmpl

		result.append(total_flux)
	return result


if __name__=="__main__":
	#print("Testing Model:",__file__)
	#from TestData import TestData
	#TestData=gradient(TestData,"rho")
	#print("Testing Result:", TestData.quantities["Gradient_rho_x1"], TestData.quantities["Gradient_rho_x2"]) #update 
	#TestData=divergence(TestData,"rho")
	#print("Testing Result:", TestData.quantities["Divergence_rho"]) #update
	''' 
	from matplotlib import pyplot as plt
	plt.figure("imshow",facecolor="lightgray")
	plt.imshow(TestData.quantities["Divergence_rho"],cmap="RdBu",extent=[TestData.grid["x1"][0],TestData.grid["x1"][-1],TestData.grid["x2"][0],TestData.grid["x2"][-1]])
	plt.colorbar()
	plt.xticks(TestData.grid["x1"])
	plt.yticks(TestData.grid["x2"])
	plt.show()
	'''
	'''
	r=np.arange(0,10,0.01)
	print(integrate_sph_body(1,r))
	'''
	print(integrate_body(np.arange(10),np.arange(0,1,0.1)))
	#f=np.full((512,256,128),(1,1,1))
	#f=np.random.random((512,256))
	#r=np.arange(2,200,198/512)
	#th=np.arange(0,np.pi,np.pi/256)
	#print(integrate_sph_body(f[:2],r[:2],th))


#File type: <Function> return np.ndarray
#By Junxiang H., 2023/07/08
#wacmk.com/cn Tech. Supp.

import numpy as np
def get(ft,unit=1,interval=1,persent=1):
	lf=np.abs(np.fft.fft(ft,n=int(persent*len(ft))))
	fre=np.fft.fftfreq(n=len(lf),d=unit*interval)
	lf=np.fft.fftshift(lf)
	fre=np.fft.fftshift(fre)
	lf=lf[int(len(lf)/2):]
	fre=fre[int(len(fre)/2):]
	return (fre,lf)

if __name__=="__main__":
	print("Testing Model:",__file__)
	from matplotlib import pyplot as plt
	t=0.001*np.arange(10000)
	lt=np.sin(2*np.pi*t)
	fre,lf=get(lt)
	plt.plot(fre,lf)
	plt.show()
	print("Testing Result:", ) #update here!


#File type: <Function>
#By Junxiang H., 2023/07/03
#wacmk.com/cn Tech. Supp.
try:
	import ShockFinder.Addon.Painter.Basic as Basic
except:
	import Basic
import pandas as pd,numpy as np,copy
def CreateSurface(**args):
	x=Basic.get_par(args,"x")
	y=Basic.get_par(args,"y")
	z=Basic.get_par(args,"z")
	v=Basic.get_par(args,"v")
	if type(x) == type(None) or type(y) == type(None) or type(v) == type(None):
		return None
	label=Basic.get_par(args,"label","")
	if type(z)!=type(None):
		surface={
			"x":x,
			"y":y,
			"v":v,
			"z":z}
	else:
		surface={
			"x":x,
			"y":y,
			"v":v}
	surfaceinfo={
			#"label":Basic.CharUndecode(label)
		}
	for i in args.keys():
		if i not in surfaceinfo.keys() and i not in surface.keys():
			surfaceinfo[i]=args[i]
	return (surface,surfaceinfo)
	

#Used to MySQL
def FormatSurface(Surface,Surfaceinfo,Surfaceid):
	newsurface=copy.deepcopy(Surface)
	if "z" in newsurface.keys():
		newsurface["v"]=[str(tuple([str(tuple(j.tolist())).replace(" ","") for j in i])).replace(" ","").replace("\"","").replace("\'","") for i in newsurface["v"] ]
	else:
		newsurface["v"]=[str(tuple(i.tolist())).replace(" ","") for i in newsurface["v"] ]
	Surfaceframe=pd.DataFrame(newsurface)
	Surfaceinfoframe=pd.DataFrame(Surfaceinfo,index=[Surfaceid])
	return (Surfaceframe,Surfaceinfoframe)

def DecodeSurface(Surfaceframe,Surfaceinfoframe):
	data=dict(zip(["x","y","v","z"],[np.array(list(Surfaceframe.to_dict()[i].values())) for i in Surfaceframe.keys()]))
	if "z" in data.keys():
		data["v"]=np.array([np.array([np.array([float(k) for k in j.replace("(","").replace(")","").split(",")]) for j in i.split("),(")]) for i in data["v"]])
	else:
		data["v"]=np.array([np.array([float(j) for j in i.replace("(","").replace(")","").split(",")]) for i in data["v"]])
	data.update(Surfaceinfoframe.to_dict())
	return CreateSurface(**data)

def info():
	print("Module:",__file__)
	
if __name__=="__main__":
	x=np.arange(0,100)
	y=np.arange(0,100)
	sur0,surinfo0=CreateSurface(x=x,y=x,z=x,v=x.reshape(len(x),1,1)*(y.reshape(1,len(y),1)-100)*x.reshape(1,1,len(x)))
	#print(sur0)
	sur0frame,surinfo0infoframe=FormatSurface(sur0,surinfo0,0)
	#print(sur0frame)
	#print(surinfo0infoframe)
	#print(DecodeSurface(sur0frame,surinfo0infoframe.loc[0])) #mysql


#File type: algroithm <Function> set
#By Junxiang H., 2023/07/02
#wacmk.com/cn Tech. Supp.

def get_par(Dataobj,vargs,quantity_name,default=None):
	if quantity_name not in Dataobj.quantities.keys() and quantity_name not in vargs.keys():
		return default
	try:
		return vargs[quantity_name]
	except:
		return Dataobj.quantities[quantity_name]


#File type: <Function> return tuple (time_index, dir, file_type)
#By Junxiang H., 2023/06/30
#wacmk.com/cn Tech. Supp.

import os
def FileNamePreProcess(filename):
	filedir,filename=os.path.split(filename)
	fileindex,filetype=os.path.splitext(filename)
	return (fileindex.split(".")[-1],filedir,filetype.split(".")[1])

if __name__=="__main__":
	print("Testing Model:",__file__)
	testvar="testfolder1/testfolder2/testfolder3/testindex.testtype"
	print("Testing Var:", testvar)
	print("Testing Result:", FileNamePreProcess(testvar))


#File type: Fixed <Class>
#By Junxiang H., 2023/07/1
#wacmk.com/cn Tech. Supp.

#Fixed files do not modify

class Data:
	def __init__(self, *args):
		self.grid=args[0]
		self.quantities=args[1]
	def update(self,quantities):
		self.quantities.update(quantities)


#File type: <Function> set
#By Junxiang H., 2023/07/03
#wacmk.com/cn Tech. Supp.



import numpy as np
import math
from scipy.interpolate import griddata

def get_par(args,name,default=None):
	try:
		return args[name]
	except:
		return default

def set_None(string):
	return None if string=="None" else string

def CharUndecode(string):
	return string.replace("$",r"$")

def find_index(x,value):
	result=np.where(x<value)[0]
	if len(result)==0:
		return 0
	elif len(x)==len(result):
		return None
	if value-x[result[-1]]>x[result[-1]+1]-value:
		return result[-1]+1
	return result[-1]

'''
def get_2drp_value(r,phi,x,y,v,default):
	tr=(x**2+y**2)**0.5
	tp=math.atan(y/x)
	if x<0:
		tp+=np.pi
	elif y<0:
		tp+=2*np.pi
	rind=find_index(r,tr)
	if rind==None:
		return default
	pind=find_index(phi,tp)
	if pind==None:
		return default
	return v[rind][pind]

def rop_to_xoy(r,phi,v=None):
	if type(v)==type(None):
		return (r*np.cos(phi),r*np.sin(phi))
	x=y=np.union1d(-r,r)
	newv=np.full((len(x),len(y)),0.5*np.min(v) if np.min(v)>0 else 1.5*np.min(v))
	newv=np.array([np.array([get_2drp_value(r,phi,x[i],y[j],v,newv[i][j]) for j in range(len(newv[i]))]) for i in range(len(newv))])
	return (x,y,newv)


def get_2drt_value(r,theta,x,z,v,default):
	tr=(x**2+z**2)**0.5
	tt=math.acos(z/tr)
	rind=find_index(r,tr)
	if rind==None:
		return default
	tind=find_index(theta,tt)
	if tind==None:
		return default
	return v[rind][tind]
def rot_to_xoz(r,theta,v=None):
	if type(v)==type(None):
		return (r*np.sin(theta),r*np.cos(theta))
	x=z=np.union1d(-r,r)
	newv=np.full((len(x),len(z)),0.5*np.min(v) if np.min(v)>0 else 1.5*np.min(v))
	newv=np.array([np.array([get_2drt_value(r,theta,x[i],z[j],v,newv[i][j]) for j in range(len(newv[i]))]) for i in range(len(newv))])
	return (x,z,newv)

def get_3drzp_value(r,phi,x,y,k,v,default):
	tr=(x**2+y**2)**0.5
	tp=math.atan(y/x)
	if x<0:
		tp+=np.pi
	elif y<0:
		tp+=2*np.pi
	rind=find_index(r,tr)
	if rind==None:
		return default
	pind=find_index(phi,tp)
	if pind==None:
		return default
	return v[rind][pind][k]

def rzp_to_xyz(r,z,phi,v=None):
	if type(v)==type(None):
		return (r*np.cos(phi),r*np.sin(phi),z)
	x=y=np.union1d(-r,r)
	newv=np.full((len(x),len(y),len(z)),0.5*np.min(v) if np.min(v)>0 else 1.5*np.min(v))
	newv=np.array([np.array([np.array([get_3drzp_value(r,phi,x[i],y[j],k,v,newv[i][j][k]) for k in range(len(newv[i][j]))]) for j in range(len(newv[i]))]) for i in range(len(newv))])
	return (x,y,z,newv)

def get_3drtp_value(r,theta,phi,x,y,z,v,default):
	tr=(x**2+y**2+z**2)**0.5
	tt=math.acos(z/tr)
	tp=math.atan(y/x)
	if x<0:
		tp+=np.pi
	elif y<0:
		tp+=2*np.pi
	rind=find_index(r,tr)
	if rind==None:
		return default
	tind=find_index(theta,tt)
	if tind==None:
		return default
	pind=find_index(phi,tp)
	if pind==None:
		return default
	
	return v[rind][tind][pind]
	
def rtp_to_xyz(r,theta,phi,v=None):
	if type(v)==type(None):
		return (r*np.sin(theta)*np.cos(phi),r*np.sin(theta)*np.sin(phi),r*np.cos(theta))
	x=y=z=np.union1d(-r,r)
	newv=np.full((len(x),len(y),len(z)),0.5*np.min(v) if np.min(v)>0 else 1.5*np.min(v))
	newv=np.array([np.array([np.array([get_3drtp_value(r,theta,phi,x[i],y[j],z[k],v,newv[i][j][k]) for k in range(len(newv[i][j]))]) for j in range(len(newv[i]))]) for i in range(len(newv))])
	return (x,y,z,newv)
'''


def pol_to_xoy(r, phi, v=None):
    # Convert the r and phi grids to 1D
    r = np.asarray(r).flatten()
    phi = np.asarray(phi).flatten()

    # Create a meshgrid for r and phi
    R, Phi = np.meshgrid(r, phi, indexing="ij")

    # Convert polar to Cartesian coordinates
    X = R * np.cos(Phi)
    Y = R * np.sin(Phi)

    # Flatten the Cartesian coordinates for griddata
    points = np.c_[X.flatten(), Y.flatten()]

    # Create a new grid in Cartesian space
    grid_x = np.linspace(-np.max(r), np.max(r), len(r))
    grid_y = np.linspace(-np.max(r), np.max(r), len(r))

    grid_X, grid_Y = np.meshgrid(grid_x, grid_y, indexing="ij")

    if v is not None:
        values = v.flatten()
        grid_V = griddata(points, values, (grid_X, grid_Y), method='linear', fill_value=0.1 * np.min(v))
        return grid_x, grid_y, grid_V

    return grid_x, grid_y

def rot_to_xoz(r,theta,v=None):
    # Convert the r and phi grids to 1D
    r = np.asarray(r).flatten()
    theta = np.asarray(theta).flatten()

    # Create a meshgrid for r and phi
    R, Th = np.meshgrid(r, theta, indexing="ij")

    # Convert polar to Cartesian coordinates
    X = R * np.sin(theta)
    Z = R * np.cos(theta)

    # Flatten the Cartesian coordinates for griddata
    points = np.c_[X.flatten(), Z.flatten()]

    # Create a new grid in Cartesian space
    grid_x = np.linspace(-np.max(r), np.max(r), len(r))
    grid_z = np.linspace(-np.max(r), np.max(r), len(r))

    grid_X, grid_Z = np.meshgrid(grid_x, grid_z, indexing="ij")

    if v is not None:
        values = v.flatten()
        grid_V = griddata(points, values, (grid_X, grid_Z), method='linear', fill_value=0.1 * np.min(v))
        return grid_x, grid_z, grid_V

    return grid_x, grid_z

def rzp_to_xyz(r, z, phi, v=None):
    # Convert the r, z, and phi grids to 1D
    r = np.asarray(r).flatten()
    z = np.asarray(z).flatten()
    phi = np.asarray(phi).flatten()

    # Create a meshgrid for r, z, and phi
    R, Z, Phi = np.meshgrid(r, z, phi, indexing="ij")

    # Convert cylindrical to Cartesian coordinates
    X = R * np.cos(Phi)
    Y = R * np.sin(Phi)

    # Flatten the Cartesian coordinates for griddata
    points = np.c_[X.flatten(), Y.flatten(), Z.flatten()]

    # Create a new grid in Cartesian space
    grid_x = np.linspace(-np.max(r), np.max(r), len(r))
    grid_y = np.linspace(-np.max(r), np.max(r), len(r))
    grid_z = np.linspace(-np.max(z), np.max(z), len(z))

    grid_X, grid_Y, grid_Z = np.meshgrid(grid_x, grid_y, grid_z, indexing="ij")

    if v is not None:
        values = v.flatten()
        grid_V = griddata(points, values, (grid_X, grid_Y, grid_Z), method='linear', fill_value=0.1 * np.min(v))
        return grid_x, grid_y, grid_z, grid_V

    return grid_x, grid_y, grid_z

def rtp_to_xyz(r, theta, phi, v=None):
    # Convert the r, theta, phi grids to 1D
    r = np.asarray(r).flatten()
    theta = np.asarray(theta).flatten()
    phi = np.asarray(phi).flatten()

    # Create a meshgrid for r, theta, phi
    R, Theta, Phi = np.meshgrid(r, theta, phi, indexing="ij")

    # Convert spherical to Cartesian coordinates
    X = R * np.sin(Theta) * np.cos(Phi)
    Y = R * np.sin(Theta) * np.sin(Phi)
    Z = R * np.cos(Theta)

    # Flatten the Cartesian coordinates for griddata
    points = np.c_[X.flatten(), Y.flatten(), Z.flatten()]

    # Create a new grid in Cartesian space
    grid_r = np.linspace(0, np.max(r), len(r))
    grid_theta = np.linspace(0, np.pi, len(theta))
    grid_phi = np.linspace(0, 2 * np.pi, len(phi))

    grid_x = np.linspace(-np.max(r), np.max(r), len(r))
    grid_y = np.linspace(-np.max(r), np.max(r), len(r))
    grid_z = np.linspace(-np.max(r), np.max(r), len(r))
    
    grid_X, grid_Y, grid_Z = np.meshgrid(grid_x, grid_y, grid_z, indexing="ij")

    if v is not None:
        values = v.flatten()
        grid_V = griddata(points, values, (grid_X, grid_Y, grid_Z), method='linear', fill_value=0.1 * np.min(v))
        return grid_X, grid_Y, grid_Z, grid_V

    return grid_x, grid_x, grid_x


def _3D_to_2D(x, y, z, v, cx=None, cy=None, cz=None):
    def closest_index(arr, value):
        """Get the index of the closest value in arr to the specified value."""
        return np.argmin(np.abs(arr - value))
    
    results = []
    
    if cx is not None:
        for c in cx:
            ix = closest_index(x, c)
            results.append((y, z, v[ix, :, :]))
            
    if cy is not None:
        for c in cy:
            iy = closest_index(y, c)
            results.append((x, z, v[:, iy, :]))
            
    if cz is not None:
        for c in cz:
            iz = closest_index(z, c)
            results.append((x, y, v[:, :, iz]))
            
    return results[0]

def info():
	print("Module:",__file__)

if __name__=="__main__":
	info()


#File type: <Function> set
#By Junxiang H., 2023/07/03
#wacmk.com/cn Tech. Supp.

try:
	import ShockFinder.Addon.Painter.Basic as Basic
except:
	import Basic

from matplotlib import pyplot as plt, colors
import pandas as pd,numpy as np
def set_figure_info(**Figureinfo):
	title=Basic.CharUndecode(Basic.get_par(Figureinfo,"title",""))
	x_axis=Basic.CharUndecode(Basic.get_par(Figureinfo,"x_axis",""))
	y_axis=Basic.CharUndecode(Basic.get_par(Figureinfo,"y_axis",""))
	xscale=Basic.get_par(Figureinfo,"xscale")
	yscale=Basic.get_par(Figureinfo,"yscale")
	plt.title(title)
	plt.xlabel(x_axis)
	plt.ylabel(y_axis)
	if xscale!=None:
		plt.xscale(xscale)
	if yscale!=None:
		plt.yscale(yscale)
	
def line(*lines,**Figureinfo):
	showlabel=False
	set_figure_info(**Figureinfo)
	x_lim=Basic.get_par(Figureinfo,"x_lim")
	y_lim=Basic.get_par(Figureinfo,"y_lim")
	if x_lim!=None:
		plt.xlim(*x_lim)
	if y_lim!=None:
		plt.ylim(*y_lim)
	for line in lines:
		le,li=line
		plt.plot(le["x"],le["y"],**li)
		if Basic.get_par(li,"label")!=None:
			showlabel=True
	if showlabel:
		plt.legend()
	plt.show()

def line_share_x(line2,*line1,**Figureinfo):
	set_figure_info(**Figureinfo)
	showlabel=False
	x_lim=Basic.get_par(Figureinfo,"x_lim")
	y_lim=Basic.get_par(Figureinfo,"y_lim")
	if x_lim!=None:
		plt.xlim(*x_lim)
	if y_lim!=None:
		plt.ylim(*y_lim)
	for line in line1:
		le,li=line
		plt.plot(le["x"],le["y"],**li)
		if Basic.get_par(li,"label")!=None:
			showlabel=True
	tx=plt.twinx()
	y2=Basic.get_par(Figureinfo,"y_axis2")
	if y2!=None:
		tx.set_ylabel(y2)
	le,li=line2
	tx.plot(le["x"],le["y"],**li)
	if Basic.get_par(li,"label")!=None:
		showlabel=True
	y_lim2=Basic.get_par(Figureinfo,"y_lim2")
	if y_lim2!=None:
		tx.set_ylim(*y_lim2)
	if showlabel:
		plt.legend()
	plt.show()

def line_share_y(line2,*line1,**Figureinfo):
	set_figure_info(**Figureinfo)
	x_lim=Basic.get_par(Figureinfo,"x_lim")
	y_lim=Basic.get_par(Figureinfo,"y_lim")
	if x_lim!=None:
		plt.xlim(*x_lim)
	if y_lim!=None:
		plt.ylim(*y_lim)
	for line in line1:
		le,li=line
		plt.plot(le["x"],le["y"],**li)
		if Basic.get_par(li,"label")!=None:
			showlabel=True
	tx=plt.twiny()
	x2=Basic.get_par(Figureinfo,"x_axis2")
	if x2!=None:
		tx.set_xlabel(x2)
	le,li=line2
	tx.plot(le["x"],le["y"],**li)
	if Basic.get_par(li,"label")!=None:
		showlabel=True
	x_lim2=Basic.get_par(Figureinfo,"x_lim2")
	if x_lim2!=None:
		tx.set_xlim(*x_lim2)
	if showlabel:
		plt.legend()
	plt.show()

def surface(surface,**Figureinfo):
	set_figure_info(**Figureinfo)
	sf,sfi=surface
	x_lim=Basic.get_par(Figureinfo,"x_lim")
	y_lim=Basic.get_par(Figureinfo,"y_lim")
	if x_lim!=None:
		plt.xticks(*x_lim)
	else:
		plt.xticks([i[int(len(i)/2)] for i in np.split(sf["x"][:10*int(len(sf["x"])/10)],10)] if len(sf["x"])>10 else sf["x"])
	if y_lim!=None:
		plt.yticks(*y_lim)
	else:
		plt.yticks([i[int(len(i)/2)] for i in np.split(sf["y"][:10*int(len(sf["y"])/10)],10)] if len(sf["y"])>10 else sf["y"])
	#plt.figure("imshow",facecolor="lightgray")
	plt.imshow(np.flipud(sf["v"].T),cmap="RdBu",extent=[sf["x"][0],sf["x"][-1],sf["y"][0],sf["y"][-1]],norm=colors.LogNorm(),**sfi)
	plt.colorbar()
	plt.show()

def info():
	print("Module:",__file__)

if __name__=="__main__":
	import Line,Surface,math
	
	x=np.arange(0,100,0.1)
	l1=Line.CreateLine(x=x,y=x**2,label="$x^2$")
	l2=Line.CreateLine(x=x,y=x**3,label="$x^3$")
	line_share_y(l2,l1,label=True,title="Test")
	
	
	r=np.arange(100)
	p=np.arange(0,(2+0.01)*np.pi,2*0.01*np.pi)
	from scipy.special import lpmv
	v=lpmv(0,3,r).reshape(100,1)*(np.cos(p)**2).reshape(1,101)
	x,y,v=Basic.rop_to_xoy(r,p,v)
	#print(v)
	sf=Surface.CreateSurface(x=x,y=y,v=v)
	surface(sf)
	


#File type: Fixed <Function> return <Object: Data>
#By Junxiang H., 2023/07/1
#wacmk.com/cn Tech. Supp.

#Fixed files do not modify

try:
	from ShockFinder.Addon.Loader.Config import Loader 
except:
	from Addon.Loader.Config import Loader 

def Load(filename,loadername):
	return Loader[loadername].load(filename)